/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QAction>
#include <QApplication>
#include <QComboBox>
#include <QFrame>
#include <QGridLayout>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QMenu>
#include <QMenuBar>
#include <QPlainTextEdit>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QStackedWidget>
#include <QStatusBar>
#include <QStringList>
#include <QToolButton>
#include <QPixmap>
#include <QIcon>
#include <QFileDialog>
#include <QWidget>
#include <QDockWidget>
#include <QToolBar>
#include <QSplitter>
#include <QDesktopWidget>
#include <QCloseEvent>
#include <QDir>
#include <QSettings>
#include <QFont>
#include <QTextCursor>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <QUrl>
#include <QMimeData>
#include <QList>
#include <QSize>
#include <cmath>
#include <assert.h>

#include "../../core/messagebox.h"
#include "../../core/expression.h"

#include "valueswidget.h"
#include "bitfieldwidget.h"
#include "ruleeditordialog.h"
#include "creategroupdialog.h"
#include "exportoptionsdialog.h"

#include "controllerparameters.h"

TControllerParameters::TControllerParameters(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QFont								bold_font;
	QFont								text_font;
	QFont								copyright_font;
	QGridLayout							*layout;
	QGridLayout							*left_layout;
	QGridLayout							*right_layout;
	QGridLayout							*dock_layout;
	QLabel								*rules_label;
	QLabel								*search_icon_label;
	QLabel								*copyright_label;
	QSizePolicy							sp_preferred_minumumexpanding(QSizePolicy::Preferred,QSizePolicy::MinimumExpanding);
	QSizePolicy							sp_preferred_maximum(QSizePolicy::Preferred,QSizePolicy::Maximum);
	QWidget								*central_widget;
	QWidget								*left_widget;
	QWidget								*right_widget;
	QDockWidget							*dock_widget;
	QWidget								*dock_widget_contents;
	QToolBar							*toolbar;
	QList<int>							splitter_sizes;
	QDir								settings_path;
	QPoint								position;
	QSize								size;
	QRect								desktop_rect;
	QRect								widget_rect;
	QString								text;
	int									index;
	
	text_font.setFamily(QStringLiteral("Courier New"));

	bold_font.setWeight(75);
	bold_font.setBold(true);
	
	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	sp_preferred_minumumexpanding.setHorizontalStretch(0);
	sp_preferred_minumumexpanding.setVerticalStretch(0);
	
	sp_preferred_maximum.setHorizontalStretch(0);
	sp_preferred_maximum.setVerticalStretch(0);
	
	// storage location
	// Win32 in directory of application
	// *nix in home folder under hidden .name folder
	
#ifdef Q_OS_WIN

	settings_path.setPath(qApp->applicationDirPath());
	
#else
	
	settings_path = QDir::home();
	
	if(!settings_path.exists(QStringLiteral(".controllerparameters")))
	{
		settings_path.mkdir(QStringLiteral(".controllerparameters"));
	}
	
	settings_path.cd(QStringLiteral(".controllerparameters"));

#endif
	
	d_settings_path = settings_path.absolutePath();
	
	if(d_settings_path.endsWith('/'))
	{
		d_settings_path.chop(1);
	}
	
	d_rules.Set_Rule_Path(d_settings_path);
	
	this->setAcceptDrops(true);
	
	d_msg_box = new TMessageBox(this);
	
	d_settings = new QSettings(d_settings_path + "/settings.ini",QSettings::IniFormat);
	
	position = d_settings->value(QStringLiteral("Mainwindow_Position"), QPoint(30,30)).toPoint();
	size = d_settings->value(QStringLiteral("Mainwindow_Size"), QSize(1008,519)).toSize();
	
	desktop_rect = QApplication::desktop()->availableGeometry();
	
	widget_rect.moveTopLeft(position);
	widget_rect.setSize(size);
	
	widget_rect = widget_rect & desktop_rect;
	
	if(widget_rect.isValid())
	{
		this->resize(widget_rect.size());
		this->move(widget_rect.topLeft());
	}
	else
	{
		this->resize(1008,519);
		this->move(QPoint(30,30));
	}
	
	
	d_status_bar = new QStatusBar(this);
	this->setStatusBar(d_status_bar);

	dock_widget = new QDockWidget(this);
	
	// maybe a bug with floatable dock widgets on OSX ?
#ifdef Q_OS_MAC
	dock_widget->setFeatures(QDockWidget::DockWidgetMovable);
#else
	dock_widget->setFeatures(QDockWidget::DockWidgetFloatable|QDockWidget::DockWidgetMovable);
#endif

	dock_widget->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea, dock_widget);
	
	this->Create_Actions();

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);

	toolbar = new QToolBar(this);
	this->addToolBar(Qt::TopToolBarArea, toolbar);
	
	toolbar->addAction(d_file_new_action);
	toolbar->addAction(d_file_open_action);
	toolbar->addAction(d_file_export_action);
	toolbar->addAction(d_file_quit_action);
	toolbar->addSeparator();
	
	layout = new QGridLayout(central_widget);
	
	d_splitter_widget = new QSplitter(central_widget);
	layout->addWidget(d_splitter_widget,0,0,1,1);
	
	left_widget = new QWidget;
	right_widget = new QWidget;

	left_layout = new QGridLayout(left_widget);
	left_layout->setContentsMargins(0, 0, 4, 0);
	
	search_icon_label = new QLabel(central_widget);
	search_icon_label->setMinimumSize(24,24);
	search_icon_label->setMaximumSize(24,24);
	search_icon_label->setAlignment(Qt::AlignCenter);
	search_icon_label->setPixmap(QPixmap(":/gui/gui_search.png"));
	left_layout->addWidget(search_icon_label,0,0,1,1);

	d_search_edit = new QLineEdit(central_widget);
	left_layout->addWidget(d_search_edit,0,1,1,1);

	d_parameter_list = new QListWidget(central_widget);
	left_layout->addWidget(d_parameter_list,1,0,1,2);
	
	d_variable_list_edit = new QPlainTextEdit(central_widget);
	d_variable_list_edit->setFont(text_font);
	d_variable_list_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_variable_list_edit->setReadOnly(true);
	left_layout->addWidget(d_variable_list_edit,2,0,1,2);
	
	d_splitter_widget->addWidget(left_widget);
	
	right_layout = new QGridLayout(right_widget);
	right_layout->setContentsMargins(4, 0, 0, 0);

	d_parameter_name_label = new QLabel(central_widget);
	d_parameter_name_label->setFont(bold_font);
	d_parameter_name_label->setMinimumSize(QSize(0, 30));
	d_parameter_name_label->setFrameShape(QFrame::StyledPanel);
	right_layout->addWidget(d_parameter_name_label,0,0,1,1);
	
	d_parameter_help_label = new QLabel(central_widget);
	d_parameter_help_label->setSizePolicy(sp_preferred_minumumexpanding);
	d_parameter_help_label->setFrameShape(QFrame::StyledPanel);
	d_parameter_help_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	d_parameter_help_label->setWordWrap(true);
	right_layout->addWidget(d_parameter_help_label,1,0,1,1);

	d_parameter_value_stack = new QStackedWidget(central_widget);
	
	d_empty_widget = new QFrame();
	d_empty_widget->setFrameShape(QFrame::StyledPanel);
	d_parameter_value_stack->addWidget(d_empty_widget);
	
	d_values_widget = new TValuesWidget();
	d_parameter_value_stack->addWidget(d_values_widget);
	
	d_bitfield_widget = new TBitfieldWidget();
	d_parameter_value_stack->addWidget(d_bitfield_widget);
	
	right_layout->addWidget(d_parameter_value_stack,2,0,1,1);
	
	d_splitter_widget->addWidget(right_widget);
	
	dock_widget_contents = new QWidget();
	dock_layout = new QGridLayout(dock_widget_contents);

	rules_label = new QLabel(dock_widget_contents);
	dock_layout->addWidget(rules_label,0,0,1,1);

	d_rules_combo = new QComboBox(dock_widget_contents);
	d_rules_combo->setSizeAdjustPolicy(QComboBox::AdjustToContents);
	dock_layout->addWidget(d_rules_combo,0,1,1,1);

	d_add_group_button = new QToolButton(dock_widget_contents);
	d_add_group_button->setIcon(QIcon(":/gui/gui_add.png"));
	d_add_group_button->setFixedSize(20,20);
	dock_layout->addWidget(d_add_group_button,0,2,1,1);

	d_remove_group_button = new QToolButton(dock_widget_contents);
	d_remove_group_button->setIcon(QIcon(":/gui/gui_remove.png"));
	d_remove_group_button->setFixedSize(20,20);
	dock_layout->addWidget(d_remove_group_button,0,3,1,1);

	d_edit_rules_button = new QToolButton(dock_widget_contents);
	d_edit_rules_button->setIcon(QIcon(":/gui/gui_edit.png"));
	d_edit_rules_button->setFixedSize(20,20);
	dock_layout->addWidget(d_edit_rules_button,0,4,1,1);

	d_execute_rules_button = new QToolButton(dock_widget_contents);
	d_execute_rules_button->setIcon(QIcon(":/gui/gui_play.png"));
	d_execute_rules_button->setFixedSize(20,20);
	dock_layout->addWidget(d_execute_rules_button,0,5,1,1);

	d_rules_result_edit = new QPlainTextEdit(dock_widget_contents);
	d_rules_result_edit->setFont(text_font);
	d_rules_result_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_rules_result_edit->setReadOnly(true);
	dock_layout->addWidget(d_rules_result_edit,1,0,1,6);
	
	dock_widget->setWidget(dock_widget_contents);
	
	
	copyright_label = new QLabel(central_widget);
	copyright_label->setFont(copyright_font);
	copyright_label->setSizePolicy(sp_preferred_maximum);
	layout->addWidget(copyright_label,1,0,1,1);

	
	// defaults
	d_parameter_value_stack->setCurrentWidget(d_empty_widget);
	d_rule_editor_dialog = 0;
	d_create_group_dialog = 0;
	d_export_options_dialog = 0;
	
	d_file_new_action->setEnabled(false);
	d_file_export_action->setEnabled(false);
	
	splitter_sizes.push_back(d_settings->value(QStringLiteral("Splitter_Size_0"), 300).toInt());
	splitter_sizes.push_back(d_settings->value(QStringLiteral("Splitter_Size_1"), 700).toInt());
	
	d_splitter_widget->setSizes(splitter_sizes);

	
	d_rules.Load_Rules();
	
	// load rule groups and select default
	d_rules_combo->addItems(d_rules.Rule_Groups());
	
	text = d_settings->value(QStringLiteral("Default_Rule_Group"), QString()).toString();
	
	if(text.length())
	{
		index = d_rules_combo->findText(text,Qt::MatchExactly);
		
		if(!(index < 0))
		{
			d_rules_combo->setCurrentIndex(index);
		}
	}

	this->setWindowTitle(QStringLiteral("Controller Parameters Version 1.0"));
	
	rules_label->setText(QStringLiteral("Rules:"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2019 Select Calibration Incorporated.  www.selectcalibration.ca"));
	
	connect(d_file_new_action,SIGNAL(triggered(bool)),this,SLOT(File_New(void)));
	connect(d_file_open_action,SIGNAL(triggered(bool)),this,SLOT(File_Load_Parameters(void)));
	connect(d_file_execute_rules_action,SIGNAL(triggered(bool)),this,SLOT(Execute_Rules(void)));
	connect(d_file_export_action,SIGNAL(triggered(bool)),this,SLOT(File_Export(void)));
	connect(d_file_quit_action,SIGNAL(triggered(bool)),this,SLOT(close(void)));
	connect(d_edit_rules_action,SIGNAL(triggered(bool)),this,SLOT(Edit_Rules(void)));
	
	connect(d_parameter_list,SIGNAL(itemSelectionChanged(void)),this,SLOT(Parameter_Selection_Changed(void)));
	connect(d_search_edit,SIGNAL(textEdited(const QString&)),this,SLOT(Search_Changed(const QString&)));
	
	connect(d_bitfield_widget,SIGNAL(Value_Changed(const int)),this,SLOT(Value_Changed(const int)));
	connect(d_values_widget,SIGNAL(Value_Changed(const int)),this,SLOT(Value_Changed(const int)));
	
	connect(d_add_group_button,SIGNAL(clicked(void)),this,SLOT(Add_Group(void)));
	connect(d_remove_group_button,SIGNAL(clicked(void)),this,SLOT(Remove_Group(void)));
	connect(d_edit_rules_button,SIGNAL(clicked(void)),this,SLOT(Edit_Rules(void)));
	connect(d_execute_rules_button,SIGNAL(clicked(void)),this,SLOT(Execute_Rules(void)));
}

TControllerParameters::~TControllerParameters(void)
{
}

void TControllerParameters::File_New(void)
{
	d_msg_box->setText("Warning");
	d_msg_box->setInformativeText("Clear loaded parameter file?");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setIcon(QMessageBox::Question);
	d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
	d_msg_box->setDefaultButton(QMessageBox::No);
	
	if(QMessageBox::No == d_msg_box->exec())
	{
		return;
	}
	
	d_parameters.Reset();
	
	d_parameter_list->clear();
	
	d_search_edit->setText(QString());
	d_active_parameter_name = QString();
	
	d_variable_list_edit->clear();
	d_rules_result_edit->clear();
	
	d_file_new_action->setEnabled(false);
	d_file_export_action->setEnabled(false);
}

void TControllerParameters::File_Load_Parameters(void)
{
	QString								file_name;
	
	file_name = QFileDialog::getOpenFileName(this,
											 QStringLiteral("Open Parameter File"),
											 QString(),
											 QStringLiteral("Parameter Files (*.asc *.xml)"));
	
	if(file_name.length())
	{
		d_parameters.Reset();
		
		if(!d_parameters.Load_Parameter_File(file_name))
		{
			d_status_bar->showMessage(QString("ERR:  %1").arg(d_parameters.Last_Error()));
			
			d_parameter_list->clear();
			
			d_search_edit->setText(QString());
			d_active_parameter_name = QString();
			
			d_variable_list_edit->clear();
			d_rules_result_edit->clear();

			d_file_new_action->setEnabled(false);
			d_file_export_action->setEnabled(false);

			return;
		}
		
		d_status_bar->showMessage(QDir::toNativeSeparators(file_name));
		
		d_parameter_list->clear();
		
		d_search_edit->setText(QString());
		d_active_parameter_name = QString();

		d_parameter_list->addItems(d_parameters.Get_Parameter_List());
		
		this->Update_Variable_List();
		d_rules_result_edit->clear();
	
		d_file_new_action->setEnabled(true);
		d_file_export_action->setEnabled(true);
	}
}

void TControllerParameters::File_Export(void)
{
	QString								file_name;
	Utilities::TExportOptions			export_options;
	
	if(!d_export_options_dialog)
	{
		d_export_options_dialog = new TExportOptionsDialog(this);
		
		export_options.export_changed_only = d_settings->value(QStringLiteral("Export_Changed_Only"), true).toBool();
		export_options.export_original_values = d_settings->value(QStringLiteral("Export_Original_Values"), true).toBool();
		export_options.export_format_match_input = d_settings->value(QStringLiteral("Export_Matches_Import"), true).toBool();
		
		d_export_options_dialog->Reset(export_options);
	}
	
	if(QDialog::Accepted == d_export_options_dialog->exec())
	{
		
		file_name = QFileDialog::getSaveFileName(this,
												 QStringLiteral("Export Changes"),
												 QString(),
												 QStringLiteral("Text file (*.txt)"));
		
		if(file_name.length())
		{
			export_options = d_export_options_dialog->Export_Options();
			
			if(!d_parameters.Write_Parameter_Data(file_name,export_options))
			{
				d_status_bar->showMessage(QString("ERR:  %1").arg(d_parameters.Last_Error()));
			}
		}
	}
}

void TControllerParameters::Parameter_Selection_Changed(void)
{
	QList<QListWidgetItem*>				sel_items;
	TParameters::TParameter				parameter;
	bool								state;
	int									count;
	int									cntr;
	
	sel_items = d_parameter_list->selectedItems();
	
	if(sel_items.size() == 1)
	{
		d_active_parameter_name = sel_items[0]->text();
		
		parameter = d_parameters.Get_Parameter(d_active_parameter_name,&state);
		
		if(state)
		{
			if(parameter.extended_name.length() > 0)
			{
				d_parameter_name_label->setText(QString("[%1]  %2 - %3").arg(parameter.index,3,10,QChar('0')).arg(parameter.name).arg(parameter.extended_name));
			}
			else
			{
				d_parameter_name_label->setText(QString("[%1]  %2").arg(parameter.index,3,10,QChar('0')).arg(parameter.name));
			}
			d_parameter_help_label->setText(parameter.description);

			switch(parameter.type)
			{
				case TParameters::TYPE_BIT_PATTERN:
					
					d_bitfield_widget->Reset();
					d_bitfield_widget->Set_Integer(parameter.values_int[0]);
					
					d_parameter_value_stack->setCurrentWidget(d_bitfield_widget);
					
					break;

				case TParameters::TYPE_INTEGER:
				case TParameters::TYPE_INTEGER_ARRAY:
					
					count = static_cast<int>(parameter.values_int.size());
					
					d_values_widget->Reset(count);
					
					for(cntr = 0;cntr < count;++cntr)
					{
						d_values_widget->Set_Integer(cntr,parameter.values_int[cntr]);
					}
					
					d_parameter_value_stack->setCurrentWidget(d_values_widget);
					
					break;
					
				case TParameters::TYPE_FLOAT:
				case TParameters::TYPE_FLOAT_ARRAY:
					
					count = static_cast<int>(parameter.values_float.size());
					
					d_values_widget->Reset(count);
					
					for(cntr = 0;cntr < count;++cntr)
					{
						d_values_widget->Set_Float(cntr,parameter.values_float[cntr]);
					}
					
					d_parameter_value_stack->setCurrentWidget(d_values_widget);
					
					break;
					
				case TParameters::TYPE_TEXT:
				case TParameters::TYPE_TEXT_ARRAY:
					
					count = static_cast<int>(parameter.values_text.size());
					
					d_values_widget->Reset(count);
					
					for(cntr = 0;cntr < count;++cntr)
					{
						d_values_widget->Set_Text(cntr,parameter.values_text[cntr]);
					}
					
					d_parameter_value_stack->setCurrentWidget(d_values_widget);
					
					break;
				
				case TParameters::TYPE_UNKNOWN:
					break;
			}
		}
	}
	else
	{
		d_active_parameter_name = QString();
		
		d_parameter_name_label->setText(QString());
		d_parameter_help_label->setText(QString());
		
		d_parameter_value_stack->setCurrentWidget(d_empty_widget);
	}
}

void TControllerParameters::Search_Changed(
	const QString						&search_text)
{
	QList<QListWidgetItem*>				items;
	
	d_parameter_list->clear();
	
	d_parameter_list->addItems(d_parameters.Get_Parameter_List(search_text));
	
	items = d_parameter_list->findItems(d_active_parameter_name,Qt::MatchExactly);
	
	if(items.size() == 1)
	{
		d_parameter_list->setCurrentItem(items[0],QItemSelectionModel::Select);
	}
	else
	{
		d_active_parameter_name = QString();
	}
}

void TControllerParameters::Value_Changed(
	const int							index)
{
	TParameters::TParameter				parameter;
	bool								state;
	
	if(d_active_parameter_name.length() > 0)
	{
		parameter = d_parameters.Get_Parameter(d_active_parameter_name,&state);
		
		if(state)
		{
			parameter.changed = true;
			
			switch(parameter.type)
			{
				case TParameters::TYPE_BIT_PATTERN:
					
					parameter.values_int[0] = d_bitfield_widget->Integer();
					
					break;
					
				case TParameters::TYPE_INTEGER:
				case TParameters::TYPE_INTEGER_ARRAY:

					parameter.values_int[index] = d_values_widget->Integer(index);
					
					break;
					
				case TParameters::TYPE_FLOAT:
				case TParameters::TYPE_FLOAT_ARRAY:
					
					parameter.values_float[index] = d_values_widget->Float(index);
					
					break;
					
				case TParameters::TYPE_TEXT:
				case TParameters::TYPE_TEXT_ARRAY:
					
					parameter.values_text[index] = d_values_widget->Text(index);
					
					break;
	
				case TParameters::TYPE_UNKNOWN:
					break;
			}
		}
		
		d_parameters.Update_Parameter(parameter);
		
		this->Update_Variable_List();
	}
}

void TControllerParameters::Add_Group(void)
{
	QString								text;
	int									index;
	
	if(!d_create_group_dialog)
	{
		d_create_group_dialog = new TCreateGroupDialog(this);
	}
	
	d_create_group_dialog->Reset(d_rules.Rule_Groups());
	
	if(QDialog::Accepted == d_create_group_dialog->exec())
	{
		text = d_create_group_dialog->Group_Name();
		
		if(d_rules.Create_Group(text))
		{
			d_rules_combo->addItem(text);
			
			index = d_rules_combo->findText(text);
			
			if(!(index < 0))
			{
				d_rules_combo->setCurrentIndex(index);
			}
		}
	}
}

void TControllerParameters::Remove_Group(void)
{
	QString								name;
	int									index;
	
	d_msg_box->setText("Warning");
	d_msg_box->setInformativeText("Delete selected group of rules?");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setIcon(QMessageBox::Question);
	d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
	d_msg_box->setDefaultButton(QMessageBox::No);
	
	if(QMessageBox::No == d_msg_box->exec())
	{
		return;
	}
	
	name = d_rules_combo->currentText();
	index = d_rules_combo->currentIndex();
	
	d_rules.Remove_Group(name);
	d_rules_combo->removeItem(index);
}

void TControllerParameters::Edit_Rules(void)
{
	if(!d_rule_editor_dialog)
	{
		d_rule_editor_dialog = new TRuleEditorDialog(this);
	}
	
	d_rule_editor_dialog->Reset(d_rules_combo->currentText(),&d_rules);
	d_rule_editor_dialog->exec();
}

void TControllerParameters::Execute_Rules(void)
{
	std::vector<TRules::TRuleItem>		rules;
	std::vector<TExpression::TVariable>	variables;
	std::vector<TRules::TRuleItem>::const_iterator rule_iter;
	std::vector<TExpression::TVariable>::const_iterator var_iter;
	TExpression							expression;
	double								left_result;
	double								right_result;
	double								right_left;
	QString								text;
	int									cursor_position;
	bool								state;
	bool								prev_state;
	QTextCursor							cursor(d_rules_result_edit->textCursor());
	static const double					ZERO_EPSILON(0.000001);
	
	d_rules_result_edit->clear();
	
	rules = d_rules.Rules(d_rules_combo->currentText());
	
	cursor_position = d_variable_list_edit->textCursor().position();
	
	// show variable list in text display
	prev_state = d_rules_result_edit->blockSignals(true);
	
	if(rules.size())
	{
		variables = d_parameters.Get_Variables();
		
		for(var_iter = variables.begin();var_iter != variables.end();++var_iter)
		{
			expression.Add_Variable((*var_iter));
		}
		
		this->Add_Rules_Text(QStringLiteral("Evaluating Rules..."));
		
		for(rule_iter = rules.begin();rule_iter != rules.end();++rule_iter)
		{
			this->Add_Rules_Text(QString("\nRule: '%1'").arg((*rule_iter).name));

			left_result = expression.Evaluate((*rule_iter).left_expression,&state);
			
			if(!state)
			{
				this->Add_Rules_Text(QStringLiteral("  ERR:  Left expression invalid."));
			}
			else
			{
				this->Add_Rules_Text(QString("  Left expression value: %1").arg(left_result,0,'f',6));
				
				right_result = expression.Evaluate((*rule_iter).right_expression,&state);

				if(!state)
				{
					this->Add_Rules_Text(QStringLiteral("  ERR:  Right expression invalid."));
				}
				else
				{
					this->Add_Rules_Text(QString("  Right expression value: %1").arg(right_result,0,'f',6));
					
					state = true;
					
					right_left = right_result - left_result;
					
					switch((*rule_iter).rule_operator)
					{
						case TRules::OPERATOR_LESS_THAN:
							
							if((right_left) < 0.0)
							{
								state = false;
							}
							
							text = QStringLiteral("  OPER:  Left Less Than Right");
							
							break;
							
						case TRules::OPEATOR_EQUAL:
							
							if(fabs(right_left) > ZERO_EPSILON)
							{
								state = false;
							}
							
							text = QStringLiteral("  OPER:  Left Equals Right");
							
							break;
							
						case TRules::OPERATOR_GREATER_THAN:
							
							if((right_left) > 0.0)
							{
								state = false;
							}
							
							text = QStringLiteral("  OPER:  Left Greater Than Right");

							break;
							
						case TRules::OPERATOR_NOT_LESS_THAN:
							
							if(!((right_left) < 0.0))
							{
								state = false;
							}
							
							text = QStringLiteral("  OPER:  Left Not Less Than Right");

							break;
							
						case TRules::OPERATOR_NOT_EQUAL:
							
							if(fabs(right_left) < ZERO_EPSILON)
							{
								state = false;
							}
							
							text = QStringLiteral("  OPER:  Left Not Equals Right");

							break;
							
						case TRules::OPERATOR_NOT_GREATER_THAN:
							
							if(!((right_left) > 0.0))
							{
								state = false;
							}
							
							text = QStringLiteral("  OPER:  Left Not Greater Than Right");

							break;
					}
					
					this->Add_Rules_Text(text);
					
					if(state)
					{
						this->Add_Rules_Text(QStringLiteral("PASS"));
					}
					else
					{
						this->Add_Rules_Text(QStringLiteral("* FAIL *"),true);
					}
				}
			}
			
			QCoreApplication::processEvents();	// can be slow if there are lots of rules
		}
	}
	
	d_rules_result_edit->blockSignals(prev_state);
	
	cursor.setPosition(cursor_position, QTextCursor::MoveAnchor);
	d_rules_result_edit->setTextCursor(cursor);

}

void TControllerParameters::closeEvent(
	QCloseEvent 						*event)
{
	Utilities::TExportOptions			export_options;
	QList<int>							splitter_sizes;

	d_rules.Save_Rules();
	
	if(d_export_options_dialog)
	{
		export_options = d_export_options_dialog->Export_Options();
		
		d_settings->setValue(QStringLiteral("Export_Changed_Only"), export_options.export_changed_only);
		d_settings->setValue(QStringLiteral("Export_Original_Values"), export_options.export_original_values);
		d_settings->setValue(QStringLiteral("Export_Matches_Import"), export_options.export_format_match_input);
	}
	
	splitter_sizes = d_splitter_widget->sizes();
	
	if(splitter_sizes.size() > 1)
	{
		d_settings->setValue(QStringLiteral("Splitter_Size_0"), splitter_sizes[0]);
		d_settings->setValue(QStringLiteral("Splitter_Size_1"), splitter_sizes[1]);
	}

	d_settings->setValue(QStringLiteral("Mainwindow_Position"), this->pos());
	d_settings->setValue(QStringLiteral("Mainwindow_Size"), this->size());

	d_settings->setValue(QStringLiteral("Default_Rule_Group"), d_rules_combo->currentText());

	
	event->accept();
}

void TControllerParameters::dragEnterEvent(
	QDragEnterEvent 					*drag_enter_event)
{
	if(drag_enter_event->mimeData()->hasFormat("text/uri-list"))
	{
		drag_enter_event->acceptProposedAction();
	}
}

void TControllerParameters::dropEvent(
	QDropEvent 							*drop_event)
{
	QList<QUrl>							urls;
	QString								file_name;
	
	urls = drop_event->mimeData()->urls();
	
	if(urls.size() == 1)
	{		
		file_name = (*urls.begin()).toLocalFile();

		if(file_name.length())
		{
			d_parameters.Reset();
			
			if(!d_parameters.Load_Parameter_File(file_name))
			{
				d_status_bar->showMessage(QString("ERR:  %1").arg(d_parameters.Last_Error()));
				
				d_parameter_list->clear();
				
				d_search_edit->setText(QString());
				d_active_parameter_name = QString();
				
				d_variable_list_edit->clear();
				d_rules_result_edit->clear();

				d_file_new_action->setEnabled(false);
				d_file_export_action->setEnabled(false);
				
				return;
			}
			
			d_status_bar->showMessage(QDir::toNativeSeparators(file_name));
			
			d_parameter_list->clear();
			
			d_search_edit->setText(QString());
			d_active_parameter_name = QString();
			
			d_parameter_list->addItems(d_parameters.Get_Parameter_List());
			
			this->Update_Variable_List();
			d_rules_result_edit->clear();

			d_file_new_action->setEnabled(true);
			d_file_export_action->setEnabled(true);
		}
	}
}

void TControllerParameters::Create_Actions(void)
{
	QMenu								*menu_Edit;
	QMenu								*menu_File;
	QMenuBar							*main_menu;

	d_file_new_action = new QAction(this);
	d_file_new_action->setToolTip(QStringLiteral("Clear parameter list."));
	d_file_new_action->setShortcut(QStringLiteral("Ctrl+N"));
	d_file_new_action->setIcon(QIcon(":/menu/menu_new.png"));
	
	d_file_open_action = new QAction(this);
	d_file_open_action->setToolTip(QStringLiteral("Open parameter file."));
	d_file_open_action->setShortcut(QStringLiteral("Ctrl+O"));
	d_file_open_action->setIcon(QIcon(":/menu/menu_open.png"));
	
	d_file_execute_rules_action = new QAction(this);
	d_file_execute_rules_action->setToolTip(QStringLiteral("Execute rules"));
	d_file_execute_rules_action->setShortcut(QStringLiteral("Ctrl+R"));
	d_file_execute_rules_action->setIcon(QIcon(":/gui/gui_play.png"));

	d_file_export_action = new QAction(this);
	d_file_export_action->setToolTip(QStringLiteral("Export changes"));
	d_file_export_action->setShortcut(QStringLiteral("Ctrl+E"));
	d_file_export_action->setIcon(QIcon(":/menu/menu_export.png"));
	
	d_file_quit_action = new QAction(this);
	d_file_quit_action->setToolTip(QStringLiteral("Close controller parameter utility."));
	d_file_quit_action->setShortcut(QStringLiteral("Ctrl+Q"));
	d_file_quit_action->setIcon(QIcon(":/menu/menu_quit.png"));
	
	d_edit_rules_action = new QAction(this);
	d_edit_rules_action->setToolTip(QStringLiteral("Edit rules"));
	d_edit_rules_action->setIcon(QIcon(":/gui/gui_edit.png"));
	
	main_menu = new QMenuBar(this);
	this->setMenuBar(main_menu);
	
	menu_File = new QMenu(main_menu);
	main_menu->addMenu(menu_File);
	menu_File->addAction(d_file_new_action);
	menu_File->addSeparator();
	menu_File->addAction(d_file_open_action);
	menu_File->addAction(d_file_execute_rules_action);
	menu_File->addAction(d_file_export_action);
	menu_File->addSeparator();
	menu_File->addAction(d_file_quit_action);
	
	menu_Edit = new QMenu(main_menu);
	main_menu->addMenu(menu_Edit);
	menu_Edit->addAction(d_edit_rules_action);

	main_menu->resize(1008,22);
	main_menu->addAction(menu_File->menuAction());
	main_menu->addAction(menu_Edit->menuAction());
	
	d_file_new_action->setText(QStringLiteral("New"));
	d_file_open_action->setText(QStringLiteral("Open"));
	d_file_execute_rules_action->setText(QStringLiteral("Execute Rules"));
	d_file_export_action->setText(QStringLiteral("Export Changes"));
	d_file_quit_action->setText(QStringLiteral("Quit"));
	d_edit_rules_action->setText(QStringLiteral("Edit Rules"));
	
	menu_File->setTitle(QStringLiteral("&File"));
	menu_Edit->setTitle(QStringLiteral("&Edit"));
}

void TControllerParameters::Update_Variable_List(void)
{
	std::vector<TExpression::TVariable>	variables;
	std::vector<TExpression::TVariable>::const_iterator var_iter;
	int									width;
	int									cursor_position;
	QString								text;
	bool								prev_state;
	QTextCursor							cursor(d_variable_list_edit->textCursor());

	
	d_variable_list_edit->clear();
	
	variables = d_parameters.Get_Variables(&width);
	
	width++;
	width *= (-1);
	
	cursor_position = d_variable_list_edit->textCursor().position();
	
	// show variable list in text display
	prev_state = d_variable_list_edit->blockSignals(true);
	
	this->Add_Variables_Text(QStringLiteral("Variable List..."));
	for(var_iter = variables.begin();var_iter != variables.end();++var_iter)
	{
		text = QString("%1 = %2").arg((*var_iter).name,width).arg((*var_iter).value,0,'f',6);
		this->Add_Variables_Text(text);
	}
	
	d_variable_list_edit->blockSignals(prev_state);
	
	cursor.setPosition(cursor_position, QTextCursor::MoveAnchor);
	d_variable_list_edit->setTextCursor(cursor);
}

void TControllerParameters::Add_Rules_Text(
	const QString						&display_text,
	const bool							bold)
{
	if(bold)
	{
		
		d_rules_result_edit->appendHtml(QString("<html><b>%1</b</html>").arg(display_text));
	}
	else
	{
		d_rules_result_edit->appendPlainText(display_text);
	}
}

void TControllerParameters::Add_Variables_Text(
	const QString						&display_text)
{
	d_variable_list_edit->appendPlainText(display_text);
}


