/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QTextStream>
#include <QFile>
#include <QByteArray>
#include <QRegExp>
#include <QStringList>
#include <QList>
#include <assert.h>

#include "../../core/xmlfile.h"

#include "utilities.h"

#include "parameters.h"

TParameters::TParameters(void)
{
}

TParameters::~TParameters(void)
{
}

QStringList TParameters::Get_Parameter_List(
	const QString						&filter) const
{
	QStringList							list;
	bool								use_filter(filter.length());
	std::vector<TParameters::TParameter>::const_iterator iter;
	
	for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
	{
		if(use_filter)
		{
			if((*iter).name.contains(filter,Qt::CaseInsensitive))
			{
				list.push_back((*iter).name);
			}
		}
		else
		{
			list.push_back((*iter).name);
		}
	}
	
	return list;
}

TParameters::TParameter TParameters::Get_Parameter(
	const QString						&name,
	bool * const						state) const
{
	std::vector<TParameters::TParameter>::const_iterator iter;
	
	if(state)
	{
		*state = false;
	}
	
	for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
	{
		if((*iter).name == name)
		{
			if(state)
			{
				*state = true;
			}

			return (*iter);
		}
	}
	
	return TParameters::TParameter();
}

std::vector<TExpression::TVariable> TParameters::Get_Variables(
	int * const							max_width) const
{
	std::vector<TParameters::TParameter>::const_iterator iter;
	std::vector<int>::const_iterator	int_iter;
	std::vector<double>::const_iterator	float_iter;
	std::vector<TExpression::TVariable>	variables;
	TExpression::TVariable				variable;
	int									index;
	int									width(0);
	
	for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
	{
		if((*iter).values_int.size() == 1)
		{
			variable.name = (*iter).name;
			variable.value = static_cast<double>((*iter).values_int[0]);
			
			if(variable.name.length() > width)
			{
				width = variable.name.length();
			}
			
			variables.push_back(variable);
		}
		else if((*iter).values_int.size() > 1)
		{
			index = 0;
			for(int_iter = (*iter).values_int.begin();int_iter != (*iter).values_int.end();++int_iter)
			{
				index++;
				
				variable.name = QString("%1.%2").arg((*iter).name).arg(index);
				variable.value = static_cast<double>(*int_iter);
				
				if(variable.name.length() > width)
				{
					width = variable.name.length();
				}

				variables.push_back(variable);
			}
		}
		else if((*iter).values_float.size() == 1)
		{
			variable.name = (*iter).name;
			variable.value = (*iter).values_float[0];
			
			if(variable.name.length() > width)
			{
				width = variable.name.length();
			}

			variables.push_back(variable);
		}
		else if((*iter).values_float.size() > 1)
		{
			index = 0;
			for(float_iter = (*iter).values_float.begin();float_iter != (*iter).values_float.end();++float_iter)
			{
				index++;
				
				variable.name = QString("%1.%2").arg((*iter).name).arg(index);
				variable.value = (*float_iter);
				
				if(variable.name.length() > width)
				{
					width = variable.name.length();
				}

				variables.push_back(variable);
			}
		}
	}
	
	if(max_width)
	{
		*max_width = width;
	}

	return variables;
}


bool TParameters::Write_Parameter_Data(
	const QString						&file_name,
	Utilities::TExportOptions			&export_options) const
{
	QFile								file(file_name);
	std::vector<TParameter>::const_iterator iter;
	std::vector<int>::const_iterator	iter_int;
	std::vector<double>::const_iterator	iter_float;
	std::vector<QString>::const_iterator iter_text;
	QString								text;
	int									index;
	bool								use_high_precision;
	
	if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		d_last_error = file.errorString();
		return false;
	}
	
	if(d_parameters_format == TParameters::FORMAT_CONSTANTS && export_options.export_format_match_input)
	{
		file.write("WARNINGS:\n");
		file.write("- Contents of this file does not included Reserved entries.\n");
		file.write("- File may contain additional data with alternate formatting.\n");
		file.write("- DO NOT SEND THIS FILE TO A CONTROLLER.\n");
		file.write("--------------------------------------------------------------\n");
		
		for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
		{
			if(export_options.export_changed_only == false || (*iter).changed == true)
			{
				text = QString("#%1 %2").arg((*iter).index,2,10,QChar(' ')).arg((*iter).extended_name);
				
				while(text.length() < 24)
				{
					text.append(' ');
				}
				text.append(':');
				
				switch((*iter).type)
				{
					case TParameters::TYPE_INTEGER:
						text.append(" Integer");
						break;
					case TParameters::TYPE_FLOAT:
						text.append(" Double");
						break;
					case TParameters::TYPE_TEXT:
						text.append(" Text");
						break;
					case TParameters::TYPE_INTEGER_ARRAY:
						text += QString(" Integer[%1]").arg((*iter).values_int.size());
						break;
					case TParameters::TYPE_FLOAT_ARRAY:
						text += QString(" Double[%1]").arg((*iter).values_float.size());
						break;
					case TParameters::TYPE_TEXT_ARRAY:
						text += QString(" Text[%1]").arg((*iter).values_text.size());
					break;
					case TParameters::TYPE_BIT_PATTERN:
						text.append(" Integer | bit pattern");
						break;
					case TParameters::TYPE_UNKNOWN:
						break;
				}
				
				while(text.length() < 64)
				{
					text.append(' ');
				}
				
				text += (*iter).name;
				file.write(QString("%1\n").arg(text).toLatin1());
				
				if(export_options.export_original_values  && (*iter).changed)
				{
					file.write("\tOriginal:\n");
					
					for(iter_int = (*iter).values_int_orig.begin();iter_int != (*iter).values_int_orig.end();++iter_int)
					{
						file.write(QString("%1\n").arg((*iter_int)).toLatin1());
					}
					
					for(iter_float = (*iter).values_float_orig.begin();iter_float != (*iter).values_float_orig.end();++iter_float)
					{
						text = QString("%1").arg((*iter_float),0,'E',14);
						text = this->Format_Value(text);
						
						file.write(QString("%1\n").arg(text).toLatin1());
					}
					
					for(iter_text = (*iter).values_text_orig.begin();iter_text != (*iter).values_text_orig.end();++iter_text)
					{
						file.write(QString("%1\n").arg((*iter_text)).toLatin1());
					}
					
					file.write("\tUpdated:\n");
				}
				
				for(iter_int = (*iter).values_int.begin();iter_int != (*iter).values_int.end();++iter_int)
				{
					file.write(QString("%1\n").arg((*iter_int)).toLatin1());
				}
				
				for(iter_float = (*iter).values_float.begin();iter_float != (*iter).values_float.end();++iter_float)
				{
					text = QString("%1").arg((*iter_float),0,'E',14);
					text = this->Format_Value(text);
					
					file.write(QString("%1\n").arg(text).toLatin1());
				}
				
				for(iter_text = (*iter).values_text.begin();iter_text != (*iter).values_text.end();++iter_text)
				{
					file.write(QString("%1\n").arg((*iter_text)).toLatin1());
				}
			}
		}
	}
	else if(d_parameters_format == TParameters::FORMAT_XML && export_options.export_format_match_input)
	{
		file.write("Parameter File List\n");
		file.write("==================================\n");
		
		for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
		{
			if(export_options.export_changed_only == false || (*iter).changed == true)
			{
				file.write(QString("\n\t<%1>\n").arg((*iter).name).toLatin1());
				file.write(QString("\n\t\t<!-- %1 -->\n").arg((*iter).description).toLatin1());
				
				if((*iter).changed)
				{
					if(export_options.export_original_values)
					{
						file.write("\n\t\t<!-- Original -->\n");
						
						if((*iter).values_text_orig.size() > 1)
						{
							index = 1;
							
							for(iter_text = (*iter).values_text_orig.begin();iter_text != (*iter).values_text_orig.end();++iter_text)
							{
								file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_text)).toLatin1());
								index++;
							}
						}
						else if((*iter).values_text_orig.size() == 1)
						{
							file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_text_orig[0]).toLatin1());
						}
						else if((*iter).values_float_orig.size() > 1)
						{
							index = 1;
							
							for(iter_float = (*iter).values_float_orig.begin();iter_float != (*iter).values_float_orig.end();++iter_float)
							{
								use_high_precision = Utilities::Use_High_Precision_Display(*iter_float);
								
								if(use_high_precision)
								{
									file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_float),0,'f',Utilities::DISPLAY_PRECISION_HIGH).toLatin1());
								}
								else
								{
									file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_float),0,'f',Utilities::DISPLAY_PRECISION_NORMAL).toLatin1());
								}
								index++;
							}
						}
						else if((*iter).values_float_orig.size() == 1)
						{
							use_high_precision = Utilities::Use_High_Precision_Display((*iter).values_float_orig[0]);
							
							if(use_high_precision)
							{
								file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_float_orig[0],0,'f',Utilities::DISPLAY_PRECISION_HIGH).toLatin1());
							}
							else
							{
								file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_float_orig[0],0,'f',Utilities::DISPLAY_PRECISION_NORMAL).toLatin1());
							}

						}
						else if((*iter).values_int_orig.size() > 1)
						{
							index = 1;
							
							for(iter_int = (*iter).values_int_orig.begin();iter_int != (*iter).values_int_orig.end();++iter_int)
							{
								file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_int)).toLatin1());
								index++;
							}
						}
						else if((*iter).values_int_orig.size() == 1)
						{
							file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_int_orig[0]).toLatin1());
						}

						file.write("\n\t\t<!-- Updated -->\n");
					}
				}
				
				if((*iter).values_text.size() > 1)
				{
					index = 1;
					
					for(iter_text = (*iter).values_text.begin();iter_text != (*iter).values_text.end();++iter_text)
					{
						file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_text)).toLatin1());
						index++;
					}
				}
				else if((*iter).values_text.size() == 1)
				{
					file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_text[0]).toLatin1());
				}
				else if((*iter).values_float.size() > 1)
				{
					index = 1;
					
					for(iter_float = (*iter).values_float.begin();iter_float != (*iter).values_float.end();++iter_float)
					{
						use_high_precision = Utilities::Use_High_Precision_Display(*iter_float);
						
						if(use_high_precision)
						{
							file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_float),0,'f',Utilities::DISPLAY_PRECISION_HIGH).toLatin1());
						}
						else
						{
							file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_float),0,'f',Utilities::DISPLAY_PRECISION_NORMAL).toLatin1());
						}
						index++;
					}
				}
				else if((*iter).values_float.size() == 1)
				{
					use_high_precision = Utilities::Use_High_Precision_Display((*iter).values_float[0]);
					
					if(use_high_precision)
					{
						file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_float[0],0,'f',Utilities::DISPLAY_PRECISION_HIGH).toLatin1());
					}
					else
					{
						file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_float[0],0,'f',Utilities::DISPLAY_PRECISION_NORMAL).toLatin1());
					}
				}
				else if((*iter).values_int.size() > 1)
				{
					index = 1;
					
					for(iter_int = (*iter).values_int.begin();iter_int != (*iter).values_int.end();++iter_int)
					{
						file.write(QString("\t\t<Value_%1>%2<Value_%1>\n").arg(index).arg((*iter_int)).toLatin1());
						index++;
					}
				}
				else if((*iter).values_int.size() == 1)
				{
					file.write(QString("\t\t<Value>%1<Value>\n").arg((*iter).values_int[0]).toLatin1());
				}
				
				file.write(QString("\n\t</%1>\n").arg((*iter).name).toLatin1());
			}
		}
	}
	else if(export_options.export_format_match_input == false)
	{
		file.write("Parameter File List\n");
		file.write("==================================\n");
		
		for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
		{
			if(export_options.export_changed_only == false || (*iter).changed == true)
			{
				if((*iter).extended_name.length() > 0)
				{
					file.write(QString("\n%1 - %2\n").arg((*iter).name).arg((*iter).extended_name).toLatin1());
				}
				else
				{
					file.write(QString("\n%1\n").arg((*iter).name).toLatin1());
				}
				
				if(export_options.export_original_values && (*iter).changed)
				{
					file.write("\tOriginal:\n");
					
					for(iter_int = (*iter).values_int_orig.begin();iter_int != (*iter).values_int_orig.end();++iter_int)
					{
						file.write(QString("%1\n").arg((*iter_int)).toLatin1());
					}
					
					for(iter_float = (*iter).values_float_orig.begin();iter_float != (*iter).values_float_orig.end();++iter_float)
					{
						text = QString("%1").arg((*iter_float),0,'E',14);
						text = this->Format_Value(text);
						
						file.write(QString("%1\n").arg(text).toLatin1());
					}
					
					for(iter_text = (*iter).values_text_orig.begin();iter_text != (*iter).values_text_orig.end();++iter_text)
					{
						file.write(QString("%1\n").arg((*iter_text)).toLatin1());
					}
					
					file.write("\tUpdated:\n");
				}
				
				for(iter_int = (*iter).values_int.begin();iter_int != (*iter).values_int.end();++iter_int)
				{
					file.write(QString("%1\n").arg((*iter_int)).toLatin1());
				}
				
				for(iter_float = (*iter).values_float.begin();iter_float != (*iter).values_float.end();++iter_float)
				{
					text = QString("%1").arg((*iter_float),0,'E',14);
					text = this->Format_Value(text);
					
					file.write(QString("%1\n").arg(text).toLatin1());
				}
				
				for(iter_text = (*iter).values_text.begin();iter_text != (*iter).values_text.end();++iter_text)
				{
					file.write(QString("%1\n").arg((*iter_text)).toLatin1());
				}
			}
		}
	}
	else
	{
		d_last_error = QString("Parameter format not set");
		return false;
	}
	
	file.close();
	
	return true;
}

void TParameters::Reset(void)
{
	d_parameters_format = TParameters::FORMAT_UNKNOWN;
	d_parameters.clear();
}

bool TParameters::Load_Parameter_File(
	const QString						&file_name)
{
	QFile								file(file_name);
	QByteArray							file_data;
	QString								init_file_data;
	
	if(file.open(QIODevice::ReadOnly | QIODevice::Text))
	{
		file_data = file.readAll();
		file.close();
	}
	else
	{
		d_last_error = QString("Cannot open file %1").arg(file_name);
		return false;
	}
	
	if(file_data.size() < 160)
	{
		d_last_error = QString("File %1 is too small to be a parameter file.").arg(file_name);
		return false;
	}
	
	init_file_data = QString::fromLatin1(file_data,160).simplified();
	
	if(init_file_data.contains(QStringLiteral("Machine Constants File"),Qt::CaseInsensitive))
	{
		d_parameters_format = TParameters::FORMAT_CONSTANTS;
		return this->Load_Constants_File(file_data);
	}
	else if(init_file_data.contains(QStringLiteral("<FDCParameterSettings>"),Qt::CaseInsensitive))
	{
		d_parameters_format = TParameters::FORMAT_XML;
		return this->Load_XML_File(file_name);
	}
	
	d_last_error = QStringLiteral("Unable to determine type of parameter file");
	return false;
}

void TParameters::Update_Parameter(
	const TParameter					&parameter)
{
	std::vector<TParameters::TParameter>::iterator iter;
	
	for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
	{
		if((*iter).name == parameter.name)
		{
			if(parameter.changed)
			{
				(*iter).changed = true;
				
				// check for non-zero original value
				if((*iter).values_int_orig.size() == 0)
				{
					(*iter).values_int_orig = (*iter).values_int;
				}
				
				if((*iter).values_float_orig.size() == 0)
				{
					(*iter).values_float_orig = (*iter).values_float;
				}
				
				if((*iter).values_text_orig.size() == 0)
				{
					(*iter).values_text_orig = (*iter).values_text;
				}
			}
			
			(*iter).values_int = parameter.values_int;
			(*iter).values_float = parameter.values_float;
			(*iter).values_text = parameter.values_text;
			
			return;
		}
	}
}

QString TParameters::Get_Constants_Description(
	const int							index) const
{
	QString								text;
		
	switch(index)
	{
		case 1:
			text += QStringLiteral("Resolution of scale following interpolation");
			break;
		case 2:
			text += QStringLiteral("Scale counting direction.  1 = normal, -1 = inverted.");
			break;
		case 3:
			text += QStringLiteral("Location of minus hardware checkpoint limit switch");
			break;
		case 4:
			text += QStringLiteral("Location of plus hardware checkpoint limit switch");
			break;
		case 5:
			text += QStringLiteral("Tolerance of hardware check-point limit switch (+/-)");
			break;
		case 6:
			text += QStringLiteral("Location of negative software end of travel");
			break;
		case 7:
			text += QStringLiteral("Location of positive software end of travel");
			break;
		case 8:
			text += QStringLiteral("Servo loop proportional gain");
			break;
		case 9:
			text += QStringLiteral("Servo loop velocity feed-forward gain");
			break;
		case 10:
			text += QStringLiteral("Wolf and Beck proportional servo gain.  Units are hecto-hz (100 hz or 0.1 khz)");
			break;
		case 11:
			text += QStringLiteral("SP600 proportional servo gain.  Units are khz.");
			break;
		case 12:
			text += QStringLiteral("Reserved");
			break;
		case 13:
			text += QStringLiteral("Servo loop type.  The sign changes the D/A polarity while the absolute value indicates the type.\n\n");
			text += QStringLiteral("0 = no axis\n");
			text += QStringLiteral("1 = Tach-type proportional plus feed-forward\n");
			text += QStringLiteral("2 = Observer with analog D/A output\n");
			text += QStringLiteral("3 = Observer with N.K. Digital PWM amplifier output\n");
			text += QStringLiteral("4 = Tach-type complete loop");
			break;
		case 14:
			text += QStringLiteral("Approximate jog speed when slow is enabled");
			break;
		case 15:
			text += QStringLiteral("Approximate jog speed when slow is not enabled");
			break;
		case 16:
			text += QStringLiteral("Used for engineering purposes");
			break;
		case 17:
			text += QStringLiteral("Machine speed driving into the axis limit switchs during homing sequence");
			break;
		case 18:
			text += QStringLiteral("Machine speed driving out of the limit switchs during homing sequence");
			break;
		case 19:
			text += QStringLiteral("Maximum and default axis speed.  Axis speed changed with MOVPAR command.");
			break;
		case 20:
			text += QStringLiteral("Maximum axis acceleration");
			break;
		case 21:
			text += QStringLiteral("Wolf and Beck probing cycle upper deflection limit. Probe reading must be between +/- this value to be valid");
			break;
		case 22:
			text += QStringLiteral("Jogbox exit timer. Timeout to switch between manual jog mode to controlled mode");
			break;
		case 23:
			text += QStringLiteral("Minimum distance to establish manual touch probe IJK direction");
			break;
		case 24:
			text += QStringLiteral("Direction of joystick axis.  1 = normal, -1 = inverted, 0 = disabled.  Index order: X,Y,Z,B,A,W");
			break;
		case 25:
			text += QStringLiteral("Source for the touch-trigger probe input:\n\n");
			text += QStringLiteral("0 = internal PI6 interface\n");
			text += QStringLiteral("1 = TTL interface\n");
			text += QStringLiteral("2 = PICS SYNC interface\n");
			text += QStringLiteral("3 = PICS TRIGGER interface\n\n");
			text += QStringLiteral("Not used on B3C/LC, B3CS, B6CS");
			break;
		case 26:
			text += QStringLiteral("Unused");
			break;
		case 27:
			text += QStringLiteral("Maximum number of touch probe re-seat attempts");
			break;
		case 28:
			text += QStringLiteral("Default SP600 Analog Probe trigger force in newtons.  Value overwritten by PRBGPA argument 3 (PRBFRC)");
			break;
		case 29:
			text += QStringLiteral("Upper limit for analogue probe deflection.  Deflection greater than this trigger an emergency");
			break;
		case 30:
			text += QStringLiteral("Nominal deflection of SP600 while scanning");
			break;
		case 31:
			text += QStringLiteral("Max permitted deviation of SP600 from parameter SCNPRB_OFF while scanning.  The limit is a +/- value.");
			break;
		case 32:
			text += QStringLiteral("Maximum acceleration during scanning");
			break;
		case 33:
			text += QStringLiteral("Maximum deceleration during scanning");
			break;
		case 34:
			text += QStringLiteral("Scan cut-plane PID element");
			break;
		case 35:
			text += QStringLiteral("Scan cut-plane PID element");
			break;
		case 36:
			text += QStringLiteral("Scan cut-plane PID element");
			break;
		case 37:
			text += QStringLiteral("Wrist type:\n\n");
			text += QStringLiteral("0 = none\n");
			text += QStringLiteral("1 = PH9\n");
			text += QStringLiteral("2 = PH10\n");
			text += QStringLiteral("3 = PH10MQ\n");
			text += QStringLiteral("8 = PHS Servo wrist\n");
			text += QStringLiteral("16 = CW43 wrist");
			break;
		case 38:
			text += QStringLiteral("Maximum allowed vector velocity.  Low power machines may not be able to drive all axis simultaneously at the maximum axis velocity");
			break;
		case 39:
			text += QStringLiteral("B0  - Invert Unlock Switches\n");
			text += QStringLiteral("B1  - Invert Limit Switches\n");
			text += QStringLiteral("B2  - External Power Supply\n");
			text += QStringLiteral("B3  - Invert Scale Error Signal\n");
			text += QStringLiteral("B4  - Invert Amplifier Enables\n");
			text += QStringLiteral("B5  - Default to SP600 Fine Mode\n");
			text += QStringLiteral("B6  - Enable Remote FeedHold\n");
			text += QStringLiteral("B7  - Enable Motion Filter (DEA_TAPS)\n");
			text += QStringLiteral("B8  - Enable Probing During Home\n");
			text += QStringLiteral("B9  - Disable Damp During Fast Jog\n");
			text += QStringLiteral("B10 - Eanble Fast Jog Before Homing\n");
			text += QStringLiteral("B11 - Enable SLOW speed\n");
			text += QStringLiteral("B12 - Invert Integrated Scale Limit Switches\n");
			text += QStringLiteral("B13 - Enable CW43L+TP200\n");
			text += QStringLiteral("B14 - Enable PRBRDV Command\n");
			text += QStringLiteral("B15 - Enable Safety Auto mode");
			break;
		case 40:
			text += QStringLiteral("Resolution of SP600 deflection signal");
			break;
		case 41:
			text += QStringLiteral("Upper limit of SP600 probe force in newtons.\n\nNote:  Overwritten by PRBATT argument 3 (PRBUTR). Possibly modified by PRBGPA argument 3 (PRBFRC) or PRBATT argument 2 (PRBLTR)");
			break;
		case 42:
			text += QStringLiteral("Lower limit of SP600 probe force in newtons.\n\nNote:  Overwritten by PRBATT argument 2 (PRBLTR).  Possibly modified by PRBGPA argument 3 (PRBFRC) or PRBATT argument 3 (PRBUTR)");
			break;
		case 43:
			text += QStringLiteral("SP600 single point routine velocity divisor.\n\nNote:  Overwritten by PRBATT argument 5 (PRBSPDRET).  PRBATT argument 5 maximum value is 99");
			break;
		case 44:
			text += QStringLiteral("SP600 single point routine delay in milliseconds");
			break;
		case 45:
			text += QStringLiteral("Wolf and Beck operating power mode:\n\n");
			text += QStringLiteral("0 = low power\n");
			text += QStringLiteral("1 = high power");
			break;
		case 46:
			text += QStringLiteral("Temp sensor Steinhart-Hart 'A' term");
			break;
		case 47:
			text += QStringLiteral("Temp sensor Steinhart-Hart 'B' term");
			break;
		case 48:
			text += QStringLiteral("Temp sensor Steinhart-Hart 'A' term");
			break;
		case 49:
			text += QStringLiteral("Temp sensor Steinhart-Hart 'B' term");
			break;
		case 50:
			text += QStringLiteral("Identification of the part sensor for the controller");
			break;
		case 51:
			text += QStringLiteral("Identification of the X axis sensors for the controller");
			break;
		case 52:
			text += QStringLiteral("Identification of the Y axis sensors for the controller");
			break;
		case 53:
			text += QStringLiteral("Identification of the Z axis sensors for the controller");
			break;
		case 54:
			text += QStringLiteral("Unused");
			break;
		case 55:
			text += QStringLiteral("Unused / prototype. Part of MEPRTTMP");
			break;
		case 56:
			text += QStringLiteral("Unused / prototype. Part of MEPRTTMP");
			break;
		case 57:
			text += QStringLiteral("Unused / prototype. Part of MEPRTTMP");
			break;
		case 58:
			text += QStringLiteral("SP600 centering contact timeout in seconds. Maximum time from prehit point until acquire part surface");
			break;
		case 59:
			text += QStringLiteral("SP600 centering settling timeout in seconds. Works with Centering Window (CTR_WIN). Max time from acquiring surface until settled");
			break;
		case 60:
			text += QStringLiteral("SP600 centering window.  The maximum allowed move distance from 4 consecutive 100 ms passes to be considered centered.\n\nNote:  Machine vibrations will affect system if this value is too low");
			break;
		case 61:
			text += QStringLiteral("Reserved");
			break;
		case 62:
			text += QStringLiteral("Reserved");
			break;
		case 63:
			text += QStringLiteral("Reserved");
			break;
		case 64:
			text += QStringLiteral("Reserved");
			break;
		case 65:
			text += QStringLiteral("Reserved");
			break;
		case 66:
			text += QStringLiteral("Nominal coefficient of thermal expansion for the X scale");
			break;
		case 67:
			text += QStringLiteral("Nominal coefficient of thermal expansion for the Y scale");
			break;
		case 68:
			text += QStringLiteral("Nominal coefficient of thermal expansion for the Z scale");
			break;
		case 69:
			text += QStringLiteral("Internal correction matrix of SP600 probe.\n\nLower matrix calibration steps using an SP600 probe:\n\n");
			text += QStringLiteral("1. Execute SET_PAR LDSP6MTX twice to zero both the PRBPIN and SP6_MTX\n");
			text += QStringLiteral("2. Perform standard probe calibration\n");
			text += QStringLiteral("3. Execute SET_PAR LDSP6MTX once.  PRBPIN matrix is zero after values are copied to SP6_MTX\n");
			text += QStringLiteral("4. Perform standard probe calibratio\n\n");
			text += QStringLiteral("Note:  The SET_PAR LDSP6MTX command does the following:\n\n");
			text += QStringLiteral("1. Copies the correction matrix from the PRBPIN 3x3 matrix to SP6_MTX\n");
			text += QStringLiteral("2. Zero's the PRBPIN 3x3 matrix");
			break;
		case 70:
			text += QStringLiteral("Prototype / Engineering use.  must be 0");
			break;
		case 71:
			text += QStringLiteral("Prototype / Engineering use.  must be 0");
			break;
		case 72:
			text += QStringLiteral("Prototype / Engineering use.  must be 0");
			break;
		case 73:
			text += QStringLiteral("Reserved");
			break;
		case 74:
			text += QStringLiteral("Reserved");
			break;
		case 75:
			text += QStringLiteral("Reserved");
			break;
		case 76:
			text += QStringLiteral("IP4 address of this controller (ie. xxx,xxx,xxx,xxx,0,0)");
			break;
		case 77:
			text += QStringLiteral("IP4 address of authorized host computer (ie. xxx,xxx,xxx,xxx,0,0)");
			break;
		case 78:
			text += QStringLiteral("Tachless 'observer loop' parameter");
			break;
		case 79:
			text += QStringLiteral("Tachless 'observer loop' parameter");
			break;
		case 80:
			text += QStringLiteral("Tachless 'observer loop' parameter");
			break;
		case 81:
			text += QStringLiteral("Tachless 'observer loop' parameter");
			break;
		case 82:
			text += QStringLiteral("Engineering use only");
			break;
		case 83:
			text += QStringLiteral("In position tolerance zone for each axis.  Machine must be within this distance prior to moving to the next target position. note: Rotary tables are special case");
			break;
		case 84:
			text += QStringLiteral("Wrist orientation. 0,0,0 uses the default where A0,B0 is pointing in -Z and A90,B0 is pointing in -Y");
			break;
		case 85:
			text += QStringLiteral("Offset between home and reported position.  Non-zero values will force an extra move during the homing sequence so that the AUTZER function ends at the new XYZ zero position");
			break;
		case 86:
			text += QStringLiteral("Constants for Renishaw probe changers.  rack type, ACR_PAR[1]:\n\n");
			text += QStringLiteral("0 = ACC2 (compatable with legacy constant files)\n");
			text += QStringLiteral("1 = ACC2\n");
			text += QStringLiteral("2 = SCR600\n");
			text += QStringLiteral("3 = SCR20\n\n");
			text += QStringLiteral("Rack orientation, ACR_PAR[2]:\n\n");
			text += QStringLiteral("0 = +X\n");
			text += QStringLiteral("1 = +Y\n");
			text += QStringLiteral("2 = -X\n");
			text += QStringLiteral("3 = -Y\n\n");
			text += QStringLiteral("Note:  Orientation is direction from port 1 to port 2\n\n");
			text += QStringLiteral("ACR_PAR[3] = ACC2 Time delay\n");
			text += QStringLiteral("ACR_PAR[4] = Wrist A angle\n");
			text += QStringLiteral("ACR_PAR[5] = Wrist B angle\n");
			text += QStringLiteral("ACR_PAR[6] = Not used\n");
			text += QStringLiteral("ACR_PAR[7] = Not used\n");
			text += QStringLiteral("ACR_PAR[8] = Not used\n");
			text += QStringLiteral("ACR_PAR[9] = Not used\n");
			break;
		case 87:
			text += QStringLiteral("SP600 manual retract distance if FINEPRB is enabled along with the auto retract manual probing cycle");
			break;
		case 88:
			text += QStringLiteral("Reserved");
			break;
		case 89:
			text += QStringLiteral("Reserved");
			break;
		case 90:
			text += QStringLiteral("Reserved");
			break;
		case 91:
			text += QStringLiteral("Reserved");
			break;
		case 92:
			text += QStringLiteral("Rotary table configuration word");
			break;
		case 93:
			text += QStringLiteral("Temperature compensation reference point in the BnS axis convention (kinematic YXZ and not XYZ)");
			break;
		case 94:
			text += QStringLiteral("Number of taps for DEA filter. Must be enabled in CMMCONFG for the filter to become active");
			break;
		case 95:
			text += QStringLiteral("Reserved");
			break;
		case 96:
			text += QStringLiteral("Total pulse response duration of the jerk limiter filter with FLY enabled. The jerk limiter of the Fly motion generator uses two cascaded moving average filters applied to the position reference.\n\n");
			text += QStringLiteral("Equivalent to the DEA smoothing frequency parameter.\n");
			text += QStringLiteral("F = 2 / Smoothing_Frequency * PI");
			break;
		case 97:
			text += QStringLiteral("Maximum deceleration for each axis axes. Used by FLY motion generator in response to a hold request. Fast braking from software emergency may be used depending on the machine type");
			break;
		case 98:
			text += QStringLiteral("Maximum deceleration for each axis by the FLY motion generator in response to the probe deflection during measurement");
			break;
		case 99:
			text += QStringLiteral("Values representing the maximum allowed distances of the FLY link between two movements:\n\n");
			text += QStringLiteral("FLY_DFLT[1] = Max fly distance at the beginning of movement\n");
			text += QStringLiteral("FLY_DFLT[2] = Max fly distance at the end of movement\n");
			text += QStringLiteral("FLY_DFLT[3] = Max fly distance at the beginning of measurement\n");
			text += QStringLiteral("FLY_DFLT[4] = Max fly distance at the beginning of measurement retract\n");
			text += QStringLiteral("FLY_DFLT[5] = Not used\n");
			text += QStringLiteral("FLY_DFLT[6] = Not used\n\n");
			text += QStringLiteral("The actual FLY distance are auto-limited by the FLY algorithm depending on current functioning conditions such as the movement length");
			break;
		case 100:
			text += QStringLiteral("PID servo loop integrative value");
			break;
		case 101:
			text += QStringLiteral("PID servo loop derivative value");
			break;
		case 102:
			text += QStringLiteral("DPID is used if set to 1 otherwise PID is used");
			break;
		case 103:
			text += QStringLiteral("D/A input value needed to move axis at the maximum speed. Defines the relationship between the output voltage of the D/A converter and the machine axis speed and is subsequently used for gain calculations\n\n");
			text += QStringLiteral("Gain = 32768 * DA_Volts / DA_Full_Scale\n\n");
			text += QStringLiteral("where:\n");
			text += QStringLiteral("DA_Full_Scale = 5v for FB-PC, B3C/LC, B3CS, B6CS\n");
			text += QStringLiteral("DA_Full_Scale = 10v for Sharpe32\n");
			break;
		case 104:
			text += QStringLiteral("Enables the static position error algorithm when the axis is stationary. If this flag is set the integrative parameter defines how fast the machine will identify the position error when stationary.  The static position error calculation is disabled when the axis is moving");
			break;
		case 105:
			text += QStringLiteral("Controls the position stability when the axis is stationary from static friction of the axis");
			break;
		case 106:
			text += QStringLiteral("Sets the insensitivity to position feedback of the servo loop when the axis is stationary");
			break;
		case 107:
			text += QStringLiteral("Feed forward corrects for expected position error in the machine during changes in acceleration.  The sources of the position errors is due to delays of the axis from the servo-amplifier behavior");
			break;
		case 108:
			text += QStringLiteral("One pole low pass filter of the servo-loop on the position feedback");
			break;
		case 109:
			text += QStringLiteral("Threshold for the maximum allowed position error of the axis.  If the position error is greater than this threshold a lag error is generated");
			break;
		case 110:
			text += QStringLiteral("Proportional component of the servo control loop while the FLY is active");
			break;
		case 111:
			text += QStringLiteral("Integrative component of the servo control loop while the FLY is active");
			break;
		case 112:
			text += QStringLiteral("B0  - Param Switch\n");
			text += QStringLiteral("B1  - Enable Software Braking\n");
			text += QStringLiteral("B2  - Enable DEA loop in open loop scanning\n");
			text += QStringLiteral("B3  - Enable 'Illegal Man-Auto change' error\n");
			text += QStringLiteral("B4  - Enable maximum allowed probing speed of 100 mm/s\n");
			text += QStringLiteral("B5  - Remove the anti-bounce filter on the probe\n");
			text += QStringLiteral("B6  - Enable the DEA reference generator with FLY off\n");
			text += QStringLiteral("B7  - Enable the DEA reference generator with joystick\n");
			text += QStringLiteral("B8  - Reduce TTP deceleration\n\n");
			text += QStringLiteral("Note:\n");
			text += QStringLiteral("B8 is in referece to TTP_DECL (item 98)\n");
			text += QStringLiteral("B5 prevents repeated point measurements on machines with high acceleration");
			break;
		case 113:
			text += QStringLiteral("Anti-collision devices installed:  1 = present, 0 = not present.");
			break;
		case 114:
			text += QStringLiteral("Distance between reference marks on the scale.  If enabled the controller expects to see a reference mark at this interval");
			break;
		case 115:
			text += QStringLiteral("Tolerance for the testing of the reference mark positions. If set to 0 the reference mark testing is disabled");
			break;
		case 116:
			text += QStringLiteral("Maximum allowed measuring speed.  If this parameter is is 0 or missing the max allowed probing velocity is 20 mm/sec.  If FLYPARSW B4 is enabled the max allowed probing velocity is 100 mm/sec");
			break;
		case 117:
			text += QStringLiteral("Observer gain when using a laser probe");
			break;
		case 118:
			text += QStringLiteral("Observer gain when using a SP600 probe");
			break;
		case 119:
			text += QStringLiteral("Configuration of the primary axis of the machine:\n\n");
			text += QStringLiteral("0 = single scale with single drive\n");
			text += QStringLiteral("1 = dual drive (hyperstatic)\n");
			text += QStringLiteral("2 = dual drive (LVDT)\n");
			text += QStringLiteral("3 = dual scale with single drive");
			break;
		case 120:
			text += QStringLiteral("Identification of slave scale for a dual drive or dual scale system.  If the rotary table is not used the slave is the 4th axis.  If the rotary table is present the rotary table will be the 4th axis and the slave scale will be the 5th axis. The primary axis of a dual drive system is always Y");
			break;
		case 121:
			text += QStringLiteral("The distance between the slave axis scale and the negative end of the X axis stroke for dual scale systems. This value is used by the dual scale algorithm");
			break;
		case 122:
			text += QStringLiteral("The distance between the master axis scale and the positive end of the X axis stroke for dual scale systems. This value is used by the dual scale algorithm");
			break;
		case 123:
			text += QStringLiteral("Position offset between the master and slave scales following the homing sequence for dual scale or dual drive systems. This value is used by both the dual scale algorithm and the dual drive control loop");
			break;
		case 124:
			text += QStringLiteral("Maximum error between the master and slave scales for a dual scale system.  Differences greater than this limit will trigger an emergency");
			break;
		case 125:
			text += QStringLiteral("For type 1 dual drive systems this defines the amount of correction to apply to the current slave position in order to compensate for differences between the master and slave clutches");
			break;
		case 126:
			text += QStringLiteral("High-pass filter on acceleration feed forward.  Enhances the acceleration feed forward performance when hardware integrative capability is present on the servo amplifier. The filter is disabled when this parameter is set to zero");
			break;
		case 127:
			text += QStringLiteral("Used by the algorithm performing the speed feed forward online identification and compensate for changes in the tachometer gain.  This algorithm is disabled if this parameter is set to zero");
			break;
		case 128:
			text += QStringLiteral("Defines part of the type of response to the CMMTYP command for handling the Quindos licence.\n\nWhen set to 1 a machine identifier of three letters in ASCII code is sent in response to the command CMMTYP.  For ASCII 'SMP' the ascii text is 0x534D50 or 5459280 in decimal.  If set to zero the CMMTYP command returns the standard XCL string of characters");
			break;
		case 129:
			text += QStringLiteral("Serial number of the machine.  This defines part of the type of response to the CMMTYP command for handling the Quindos licence.  With the serial number set to zero the response to the CMMTYP command will include 604 for the serial number section");
			break;
		case 130:
			text += QStringLiteral("Defines the home sequence of linear axes:\n\n");
			text += QStringLiteral("0: Vertical arm CMM, Z first, X,Y second\n");
			text += QStringLiteral("1: Vertical arm CMM, Z first, X second, Y third\n");
			text += QStringLiteral("2: Horizontal arm CMM, Y first, X,Z second\n");
			text += QStringLiteral("3: Horizontal arm CMM, Y first, X second, Z third\n");
			text += QStringLiteral("4: Horizontal arm CMM, Y first, Z second, X third");
			break;
		case 131:
			text += QStringLiteral("Describes the XYZ coordinates for the rotary table origin from a BnS compensation map");
			break;
		case 132:
			text += QStringLiteral("Theoretical width for the scale zero pulse.  This value is used by the scale auto adjustment algorithm");
			break;
		case 133:
			text += QStringLiteral("Overspeed threshold expressed as a percentage of the maximum CNC speed. If the difference between the speed command and the measured speed exceeds the threshold an emergency is generated. The theoretical threshold is used when the axis is stationary but increases by a factor of four when moving");
			break;
		case 134:
			text += QStringLiteral("Identification of the automatic adjustment file by number.  For example, 305 identifies the name of the adjustment file as ADJ305.ADJ. The parameter may be zero if no automatic adjustment file exists");
			break;
		case 135:
			text += QStringLiteral("Enable or disable control of the ENABLE CYCLE device on the FDG cells. This parameter is only enabled for those FDG cells that actually have the device.  Enabled = 1, disabled = 0.");
			break;
		case 136:
			text += QStringLiteral("Speed reduction for B3CS controllers. If enabled by setting CMMCONFG B11 to 1 the motion speed is limited to this value when a dedicated input is not signaled (0) and moves normally when the input is signaled (1)");
			break;
		case 137:
			text += QStringLiteral("Stroke reduction for B3CS controllers. Valid range [-64...+64]. If zero the stroke reduction is not enabled for PLC inputs. If not 0 the absolute value indicates the number of the PLC input used for the stroke reduction.  The sign indicates the signal levels according the following rule:\n\n");
			text += QStringLiteral("• +: reduced stroke for input at logical 0, full stroke for input at logical 1\n");
			text += QStringLiteral("• -: reduced stroke for input at logical 1, full stroke for input at logical 0\n\n");
			text += QStringLiteral("Example 1: IN_RED =  25: logical 0 on input 25 causes stroke reduction\n");
			text += QStringLiteral("Example 2: IN_RED = -14: logical 1 on input 14 causes stroke reduction");
			break;
		case 138:
			text += QStringLiteral("Stroke reduction for B3CS controllers. Valid range [-64...+64]. If zero the stroke reduction is not enabled for PLC outputs. If not 0 the absolute value indicates the number of the PLC output used for the stroke reduction.  The sign indicates the signal levels according the following rule:\n\n");
			text += QStringLiteral("• +: reduced stroke for output at logical 0, full stroke for output at logical 1\n");
			text += QStringLiteral("• -: reduced stroke for output at logical 1, full stroke for output at logical 0\n\n");
			text += QStringLiteral("Example 1: IN_RED = 26: logical 0 on output 26 causes stroke reduction\n");
			text += QStringLiteral("Example 2: IN_RED = -15: logical 1 on output 15 causes stroke reduction");
			break;
		case 139:
			text += QStringLiteral("Stroke reduction for B3CS controllers. Valid range is 0 to some positive value.  When the stroke reduction is active due to the status of PLC I/O indicated by IN_RED and OUT_RED the minimum allowed coordinate for each axis becomes M_SWLIM + M_SWLIMR.  If M_SWLIMR is 0 or negative the stroke is not reduced\n\n");
			text += QStringLiteral("Example 1: M_SWLIM = -50 mm and M_SWLIMR=30 mm.  With active reduction the minimum coordinate is -20 mm\n");
			text += QStringLiteral("Example 2: M_SWLIM = -50 mm and M_SWLIMR=70 mm.  With active reduction the minimum coordinate is  20 mm");
			break;
		case 140:
			text += QStringLiteral("Stroke reduction for B3CS controllers. Valid range is 0 to some positive value.  When the stroke reduction is active due to the status of PLC I/O indicated by IN_RED and OUT_RED the minimum allowed coordinate for each axis becomes P_SWLIM - P_SWLIMR.  If P_SWLIMR is 0 or negative the stroke is not reduced\n\n");
			text += QStringLiteral("Example: P_SWLIM = 800 mm and P_SWLIMR = 50 mm.  With active reduction the maximum coordinate is 750 mm");
			break;
		case 142:
			text += QStringLiteral("B0  - Use Polynomial Upper Level\n");
			text += QStringLiteral("B1  - Use Polynomial Lower Level\n");
			text += QStringLiteral("B2  - Use NLPPAR AGA Deadband\n");
			text += QStringLiteral("B3  - Use SP600/SP25 Oscil. Test\n");
			text += QStringLiteral("B4  - Reserved\n");
			text += QStringLiteral("B5  - Use SP600/SP25 SW Damping\n");
			text += QStringLiteral("B6  - LSPX5 Changable Temp Sensor\n");
			break;
			
		case 146:
			text += QStringLiteral("B0  - Wrist has Battery Backup\n");
			text += QStringLiteral("B1  - Unused\n");
			text += QStringLiteral("B2  - Use Tesastar-M WD Command\n");
			break;
			
		case 150:
			text += QStringLiteral("B0  - Invert Unlock Switches\n");
			text += QStringLiteral("B1  - Invert Limit Switches\n");
			text += QStringLiteral("B2  - External Power Supply\n");
			text += QStringLiteral("B3  - Invert Scale Error Signals\n");
			text += QStringLiteral("B4  - Invert Amplifier Enables\n");
			text += QStringLiteral("B5  - Default to SP600 Fine Mode\n");
			text += QStringLiteral("B6  - Enable Remote Monitor\n");
			text += QStringLiteral("B7  - Enable Motion Filter\n");
			text += QStringLiteral("B8  - Enable Probe During Home\n");
			text += QStringLiteral("B9  - Disable Damp During Fast Jog\n");
			text += QStringLiteral("B10 - Enable Fast Jog before autzer\n");
			text += QStringLiteral("B11 - Enable safety speed\n");
			text += QStringLiteral("B12 - Inv. Integrated Limit Switch\n");
			text += QStringLiteral("B13 - Enable CW43 + TP200\n");
			text += QStringLiteral("B14 - Enable PRBRDV Command\n");
			text += QStringLiteral("B15 - Enable Safety Auto mode\n");
			break;

		default:
			text += QStringLiteral("No information available");
			break;
	}
	
	return text;
}

QString TParameters::Format_Value(
	const QString						&in) const
{
	QString								out(in);
	QString								exponent;
	int									index;
	int									value;
	
	if(in.length() == 0)
	{
		return out;
	}
	
	if(out.at(0) != QChar('-'))
	{
		// add leading +
		out.insert(0,QChar('+'));
	}
	
	index = out.indexOf('E');
	
	if(!(index < 0))
	{
		exponent = out.mid(index + 1);
		out.truncate(index);
		
		value = exponent.toInt();
		
		if(value < 0)
		{
			out += QString("E-%1").arg(abs(value),4,10,QChar('0'));
		}
		else
		{
			out += QString("E+%1").arg(value,4,10,QChar('0'));
		}
	}
	
	return out;
}

TParameters::TParameter TParameters::Convert_From_Text(
	const TParameters::TParameter		&parameter) const
{
	std::vector<QString>::const_iterator iter;
	int									ival;
	double								dval;
	bool								state;
	TParameters::TParameter				new_parameter(parameter);
	
	if(!(parameter.type == TParameters::TYPE_TEXT ||
	   parameter.type == TParameters::TYPE_TEXT_ARRAY))
	{
		return new_parameter;
	}
	
	for(iter = new_parameter.values_text.begin();iter != new_parameter.values_text.end();++iter)
	{
		ival = (*iter).toInt(&state);
		
		if(state)
		{
			new_parameter.values_int.push_back(ival);
		}
		else
		{
			dval = (*iter).toDouble(&state);
			
			if(state)
			{
				new_parameter.values_float.push_back(dval);
			}
		}
	}
	
	if(new_parameter.values_int.size() == new_parameter.values_text.size())
	{
		new_parameter.values_text.clear();
		new_parameter.values_float.clear();
		
		new_parameter.values_text_orig.clear();
		new_parameter.values_float_orig.clear();
		
		new_parameter.values_int_orig = new_parameter.values_int;
		
		if(new_parameter.values_int.size() == 1)
		{
			new_parameter.type = TParameters::TYPE_INTEGER;
		}
		else
		{
			new_parameter.type = TParameters::TYPE_INTEGER_ARRAY;
		}
	}
	else if(new_parameter.values_float.size() == new_parameter.values_text.size())
	{
		new_parameter.values_text.clear();
		new_parameter.values_int.clear();
		
		new_parameter.values_text_orig.clear();
		new_parameter.values_int_orig.clear();
		
		new_parameter.values_float_orig = new_parameter.values_float;

		if(new_parameter.values_float.size() == 1)
		{
			new_parameter.type = TParameters::TYPE_FLOAT;
		}
		else
		{
			new_parameter.type = TParameters::TYPE_FLOAT_ARRAY;
		}
	}
	
	return new_parameter;
}

bool TParameters::Load_Constants_File(
	const QByteArray					&file_data)
{
	QString								line;
	QStringList							line_items;
	QString								text;
	QRegExp								rx("^[#]([ 0-9]{1,5})[ ]");
	TParameter							parameter;
	bool								parameter_active(false);
	bool								state;
	double								dval;
	int									ival;
	int									index;
	
	// defaults
	parameter.index = 0;
	
	QTextStream							in(file_data);
	
	while(!in.atEnd())
	{
		line = in.readLine().trimmed();
		
		if(!(rx.indexIn(line) < 0))
		{
			if(rx.captureCount() == 1)
			{
				// add existing parameter is it has data
				if(parameter.values_int.size() || parameter.values_float.size())
				{
					this->Add_Parameter(parameter);
				}
				
				parameter.values_int.clear();
				parameter.values_float.clear();
				parameter.values_text.clear();
				
				parameter.values_int_orig.clear();
				parameter.values_float_orig.clear();
				parameter.values_text_orig.clear();
				
				parameter.extended_name = QString();
				parameter.changed = false;
	
				parameter.index = rx.cap(1).toInt();
				parameter.name = line.mid(64);
				parameter.description = this->Get_Constants_Description(parameter.index);
				
				line.truncate(64);	// remove name from end of string
				
				index = rx.matchedLength();
				line = line.mid(index);
				
				index = line.indexOf(':');
				
				if(!(index < 0))
				{
					// name cannot be less than 20 chars
					// if it is get the next one
					while(index < 19)
					{
						index = line.indexOf(':',index + 1);
						
						if(index < 0)
						{
							index = 19;
						}
					}
					
					text = line.mid(0,index);
					
					parameter.extended_name = text.trimmed();
					
					line = line.mid(index+1).trimmed();

					parameter.type = TParameters::TYPE_UNKNOWN;
					
					if(line.contains(QStringLiteral("bit pattern"),Qt::CaseInsensitive))
					{
						parameter.type = TParameters::TYPE_BIT_PATTERN;
					}
					else
					{
						if(line.startsWith(QStringLiteral("Integer["),Qt::CaseInsensitive))
						{
							parameter.type = TParameters::TYPE_INTEGER_ARRAY;
						}
						else if(line.startsWith(QStringLiteral("Double["),Qt::CaseInsensitive))
						{
							parameter.type = TParameters::TYPE_FLOAT_ARRAY;
						}
						else if(line.startsWith(QStringLiteral("Longint["),Qt::CaseInsensitive))
						{
							parameter.type = TParameters::TYPE_INTEGER_ARRAY;
						}
						else if(line.startsWith(QStringLiteral("Longint"),Qt::CaseInsensitive))
						{
							parameter.type = TParameters::TYPE_INTEGER;
						}
						else if(line.startsWith(QStringLiteral("Integer"),Qt::CaseInsensitive))
						{
							parameter.type = TParameters::TYPE_INTEGER;
						}
						else if(line.startsWith(QStringLiteral("Double"),Qt::CaseInsensitive))
						{
							parameter.type = TParameters::TYPE_FLOAT;
						}
					}
					
					parameter_active = (parameter.type != TParameters::TYPE_UNKNOWN);
				}
			}
		}
		else
		{
			if(parameter_active)
			{
				switch(parameter.type)
				{
					case TParameters::TYPE_INTEGER:
					case TParameters::TYPE_INTEGER_ARRAY:
					case TParameters::TYPE_BIT_PATTERN:
						ival = line.toInt(&state);
						
						if(state)
						{
							parameter.values_int.push_back(ival);
						}
						else
						{
							parameter_active = false;
						}
						
						break;
						
					case TParameters::TYPE_FLOAT:
					case TParameters::TYPE_FLOAT_ARRAY:
						dval = line.toDouble(&state);
						
						if(state)
						{
							parameter.values_float.push_back(dval);
						}
						else
						{
							parameter_active = false;
						}
						
						break;
						
					case TParameters::TYPE_TEXT:
					case TParameters::TYPE_TEXT_ARRAY:
					case TParameters::TYPE_UNKNOWN:
						parameter_active = false;
						break;
				}
			}
		}
	}
	
	// add last parameter is it has data
	if(parameter.values_int.size() || parameter.values_float.size())
	{
		this->Add_Parameter(parameter);
	}
	
	return true;
}

bool TParameters::Load_XML_File(
	const QString						&file_name)
{
	TXmlFile							xml_file;
	QDomNode							node_lev_a;
	QDomNode							node_lev_b;
	QDomElement							node_lev_c;
	QDomComment							node_comment;
	TParameter							parameter;
	QString								text;
	std::vector<QString>				text_array;
	int									index(1);

	try
	{
		xml_file.SetFileName(file_name);
		xml_file.Open("FDCParameterSettings");
		
		parameter.index = index;
		parameter.changed = false;
		parameter.type = TParameters::TYPE_TEXT;
		
		parameter.values_int_orig.clear();
		parameter.values_float_orig.clear();
		parameter.values_text_orig.clear();
		
		node_lev_a = xml_file.GetFirstNode(xml_file.RootNode());
		
		while(!node_lev_a.isNull())
		{
			parameter.name = node_lev_a.nodeName();
			parameter.extended_name = QString();
			
			text_array.clear();
			
			node_lev_b = xml_file.GetFirstNode(&node_lev_a,true);
			
			while(!node_lev_b.isNull())
			{
				if(!node_lev_b.isText())
				{
					if(node_lev_b.isComment())
					{
						node_comment = node_lev_b.toComment();
						parameter.description = node_comment.nodeValue().trimmed();
					}
					else
					{
						text_array.push_back(xml_file.ReadTextNode(&node_lev_b,true));
					}
				}
				
				node_lev_b = xml_file.GetSiblingNode(&node_lev_b,true);

			};
			
			if(text_array.size())
			{
				parameter.values_text = text_array;
				parameter.index = index++;
				
				this->Add_Parameter(this->Convert_From_Text(parameter));
			}
			
			node_lev_a = xml_file.GetSiblingNode(&node_lev_a,true);

		};
	}
	catch(const TXmlFileException &xml_file_exception)
	{
		d_last_error = xml_file_exception.ErrorText();
		return false;
	}
	
	return true;
}


bool TParameters::Add_Parameter(
	const TParameter					&parameter)
{
	std::vector<TParameters::TParameter>::const_iterator iter;
	
	// don't add reserved parameters
	if(parameter.name.startsWith("--------"))
	{
		return false;
	}
	
	// check for exsisting name
	for(iter = d_parameters.begin();iter != d_parameters.end();++iter)
	{
		if(parameter.name == (*iter).name)
		{
			return false;
		}
	}
		
	d_parameters.push_back(parameter);

	return true;
}
