/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QListWidget>
#include <QPushButton>
#include <QSpacerItem>
#include <QVBoxLayout>
#include <QList>
#include <QListWidgetItem>
#include <assert.h>

#include "../../core/messagebox.h"

#include "rules.h"
#include "ruleitemdialog.h"
#include "copyrulesdialog.h"

#include "ruleeditordialog.h"

TRuleEditorDialog::TRuleEditorDialog(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*separator_hline;
	QGridLayout							*dialog_layout;
	QHBoxLayout							*button_hlayout;
	QPushButton							*close_button;
	QSpacerItem							*button_hspacer;
	QSpacerItem							*control_vspacer;
	QVBoxLayout							*control_vlayout;


	this->resize(471,392);
	
	d_msg_box = new TMessageBox(this);

	dialog_layout = new QGridLayout(this);

	control_vlayout = new QVBoxLayout();

	d_add_rule_button = new QPushButton(this);
	d_add_rule_button->setAutoDefault(false);
	control_vlayout->addWidget(d_add_rule_button);

	d_edit_rule_button = new QPushButton(this);
	d_edit_rule_button->setAutoDefault(false);
	control_vlayout->addWidget(d_edit_rule_button);

	d_delete_rule_button = new QPushButton(this);
	d_delete_rule_button->setAutoDefault(false);
	control_vlayout->addWidget(d_delete_rule_button);

	d_copy_rule_button = new QPushButton(this);
	d_copy_rule_button->setAutoDefault(false);
	control_vlayout->addWidget(d_copy_rule_button);

	control_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	control_vlayout->addItem(control_vspacer);
	dialog_layout->addLayout(control_vlayout,1,0,1,1);

	d_rule_list = new QListWidget(this);
	d_rule_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	d_rule_list->setSortingEnabled(true);
	dialog_layout->addWidget(d_rule_list,1,1,1,1);

	button_hlayout = new QHBoxLayout();

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	close_button = new QPushButton(this);
	close_button->setAutoDefault(false);
	button_hlayout->addWidget(close_button);
	dialog_layout->addLayout(button_hlayout,3,0,1,2);

	separator_hline = new QFrame(this);
	separator_hline->setFrameShape(QFrame::HLine);
	separator_hline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_hline,2,0,1,2);

	d_rule_group_label = new QLabel(this);
	dialog_layout->addWidget(d_rule_group_label,0,1,1,1);
	
	d_rule_item_dialog = new TRuleItemDialog(this);
	
	
	// defaults
	d_copy_rules_dialog = 0;

	this->setWindowTitle(QStringLiteral("Rule Editor"));
	
	d_add_rule_button->setText(QStringLiteral("Add Rule"));
	d_edit_rule_button->setText(QStringLiteral("Edit Rule"));
	d_delete_rule_button->setText(QStringLiteral("Delete Rule"));
	d_copy_rule_button->setText(QStringLiteral("Copy Rule"));
	close_button->setText(QStringLiteral("Close"));
	
	connect(close_button,SIGNAL(clicked(void)),this,SLOT(close(void)));
	connect(d_add_rule_button,SIGNAL(clicked(void)),this,SLOT(Add_Rule(void)));
	connect(d_edit_rule_button,SIGNAL(clicked(void)),this,SLOT(Edit_Rule(void)));
	connect(d_delete_rule_button,SIGNAL(clicked(void)),this,SLOT(Delete_Rule(void)));
	connect(d_copy_rule_button,SIGNAL(clicked(void)),this,SLOT(Copy_Rule(void)));

	connect(d_rule_list,SIGNAL(itemSelectionChanged(void)),this,SLOT(Rule_Selection_Changed(void)));
	connect(d_rule_list,SIGNAL(itemDoubleClicked(QListWidgetItem*)),this,SLOT(Rule_Double_Clicked(QListWidgetItem*)));
	
}

TRuleEditorDialog::~TRuleEditorDialog(void)
{
}

void TRuleEditorDialog::Reset(
	const QString						&rule_group,
	TRules * const						rules)
{
	assert(rules);
	
	d_rule_group = rule_group;
	d_rule_group_label->setText(d_rule_group);
	d_rules = rules;
	
	// update rule list for associated group
	d_rule_list->clear();
	d_rule_list->addItems(d_rules->Rule_List(rule_group));
	
	d_add_rule_button->setEnabled(true);
	d_edit_rule_button->setEnabled(false);
	d_delete_rule_button->setEnabled(false);
	d_copy_rule_button->setEnabled(false);
}

void TRuleEditorDialog::Rule_Selection_Changed(void)
{
	QList<QListWidgetItem*>				sel_items;
	
	sel_items = d_rule_list->selectedItems();
	
	if(!sel_items.size())
	{
		d_edit_rule_button->setEnabled(false);
		d_delete_rule_button->setEnabled(false);
		d_copy_rule_button->setEnabled(false);
	}
	else if(sel_items.size() == 1)
	{
		d_edit_rule_button->setEnabled(true);
		d_delete_rule_button->setEnabled(true);
		d_copy_rule_button->setEnabled(true);
	}
	else if(sel_items.size() > 1)
	{
		d_edit_rule_button->setEnabled(false);
		d_delete_rule_button->setEnabled(true);
		d_copy_rule_button->setEnabled(true);
	}
}

void TRuleEditorDialog::Rule_Double_Clicked(
	QListWidgetItem						*item)
{
	if(item)
	{
		this->Edit_Rule();
	}
}

void TRuleEditorDialog::Add_Rule(void)
{
	TRules::TRuleItem					item;
	QStringList							existing_item_rules;
	
	existing_item_rules = d_rules->Rule_List(d_rule_group);
	
	d_rule_item_dialog->Reset(TRuleItemDialog::MODE_ADD,item,existing_item_rules);
	
	if(QDialog::Accepted == d_rule_item_dialog->exec())
	{
		item = d_rule_item_dialog->Rule_Item();
		
		if(!d_rules->Add_Rule(d_rule_group,item))
		{
			d_msg_box->setText("Error");
			d_msg_box->setInformativeText("Cannot add new rule entry.");
			d_msg_box->setDetailedText(d_rules->Last_Error());
			d_msg_box->setIcon(QMessageBox::Warning);
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			
			d_msg_box->exec();
			return;
		}

		d_rule_list->addItem(item.name);
	}
}

void TRuleEditorDialog::Edit_Rule(void)
{
	TRules::TRuleItem					item;
	QListWidgetItem						*list_item;
	QStringList							existing_item_rules;
	QString								rule_name;
	bool								valid;
	
	
	existing_item_rules = d_rules->Rule_List(d_rule_group);
	
	list_item = d_rule_list->currentItem();
	
	if(!list_item)
	{
		d_msg_box->setText("Error");
		d_msg_box->setInformativeText("Cannot edit existing rule.");
		d_msg_box->setDetailedText(QStringLiteral("No rule selected"));
		d_msg_box->setIcon(QMessageBox::Warning);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
		return;
	}
	
	item = d_rules->Rule(d_rule_group,list_item->text(),&valid);
	
	if(!valid)
	{
		d_msg_box->setText("Error");
		d_msg_box->setInformativeText("Cannot edit existing rule.");
		d_msg_box->setDetailedText(d_rules->Last_Error());
		d_msg_box->setIcon(QMessageBox::Warning);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
		return;
	}
	
	rule_name = item.name;
	d_rule_item_dialog->Reset(TRuleItemDialog::MODE_EDIT,item,existing_item_rules);
	
	if(QDialog::Accepted == d_rule_item_dialog->exec())
	{
		item = d_rule_item_dialog->Rule_Item();
		
		d_rule_list->takeItem(d_rule_list->row(list_item));
		delete list_item;
		
		d_rules->Delete_Rule(d_rule_group,rule_name);
		
		if(!d_rules->Add_Rule(d_rule_group,item))
		{
			d_msg_box->setText("Error");
			d_msg_box->setInformativeText("Cannot add new rule entry.");
			d_msg_box->setDetailedText(d_rules->Last_Error());
			d_msg_box->setIcon(QMessageBox::Warning);
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			
			d_msg_box->exec();
			return;
		}
		
		list_item = new QListWidgetItem();
		list_item->setText(item.name);
		
		d_rule_list->addItem(list_item);
		d_rule_list->setCurrentItem(list_item,QItemSelectionModel::ClearAndSelect);
	}
}

void TRuleEditorDialog::Delete_Rule(void)
{
	QList<QListWidgetItem*>				sel_items;
	QList<QListWidgetItem*>::const_iterator iter;
	
	sel_items = d_rule_list->selectedItems();
	
	if(sel_items.size())
	{
		d_msg_box->setText("Warning");
		d_msg_box->setInformativeText("Delete selected rules?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setIcon(QMessageBox::Question);
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		
		if(QMessageBox::No == d_msg_box->exec())
		{
			return;
		}
	}

	for(iter = sel_items.begin();iter != sel_items.end();++iter)
	{
		d_rules->Delete_Rule(d_rule_group,(*iter)->text());
		
		d_rule_list->takeItem(d_rule_list->row(*iter));
		delete (*iter);
	}
}

void TRuleEditorDialog::Copy_Rule(void)
{
	QStringList							groups;
	QList<QListWidgetItem*>				sel_items;
	QList<QListWidgetItem*>::iterator	sel_items_iter;
	QString								target_group;
	bool								overwrite_rule_item;
	
	if(!d_copy_rules_dialog)
	{
		d_copy_rules_dialog = new TCopyRulesDialog(this);
	}
	
	groups = d_rules->Rule_Groups();
	
	if(groups.size() > 1)
	{
		groups.removeOne(d_rule_group);

		d_copy_rules_dialog->Reset(groups);

		if(QDialog::Accepted == d_copy_rules_dialog->exec())
		{
			d_msg_box->setText("Overwrite Existing Rules");
			d_msg_box->setInformativeText("Do you want to overwrite existing rules in target group (if they exist)?");
			d_msg_box->setDetailedText(QString());
			d_msg_box->setIcon(QMessageBox::Question);
			d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
			d_msg_box->setDefaultButton(QMessageBox::No);
			
			if(QMessageBox::No == d_msg_box->exec())
			{
				overwrite_rule_item = false;
			}
			else
			{
				overwrite_rule_item = true;
			}

			target_group = d_copy_rules_dialog->Target_Group();
			
			sel_items = d_rule_list->selectedItems();
			
			if(sel_items.size() > 0)
			{
				for(sel_items_iter = sel_items.begin();sel_items_iter != sel_items.end();++sel_items_iter)
				{
					d_rules->Copy_Rule(d_rule_group,target_group,(*sel_items_iter)->text(),overwrite_rule_item);
				}
			}
		}
	}
}

