/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <algorithm>
#include <cmath>
#include <assert.h>

#include "../../core/xmlfile.h"

#include "rules.h"

// sort rule items by name
struct TRuleItemSort
{
	bool operator() (TRules::TRuleItem a,TRules::TRuleItem b)
	{
		return (a.name < b.name);
	}
};


TRules::TRules(void)
{
	d_data_changed = false;
}

TRules::~TRules(void)
{
}

bool TRules::Save_Rules(void) const
{
	if(d_data_changed)
	{
		if(this->Write())
		{
			d_data_changed = false;
			return true;
		}
		
		return false;
	}
	
	return true;
}

QStringList TRules::Rule_Groups(void) const
{
	QStringList							list;
	std::vector<TRules::TRule>::const_iterator iter;
	
	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		list.push_back((*iter).name);
	}
	
	return list;
}

QStringList TRules::Rule_List(
	const QString						&rule_group) const
{
	QStringList							list;
	std::vector<TRules::TRule>::const_iterator iter;
	std::vector<TRules::TRuleItem>::const_iterator item_iter;
	
	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if(rule_group == (*iter).name)
		{
			for(item_iter = (*iter).rule_items.begin();item_iter != (*iter).rule_items.end();++item_iter)
			{
				list.push_back((*item_iter).name);
			}
			break;
		}
	}
	
	return list;
}

TRules::TRuleItem TRules::Rule(
	const QString						&rule_group,
	const QString						&rule_name,
	bool								*valid) const
{
	TRules::TRuleItem					item;
	std::vector<TRules::TRule>::const_iterator iter;
	std::vector<TRules::TRuleItem>::const_iterator item_iter;
	
	if(valid)
	{
		*valid = false;
	}
	
	item.name = rule_name;

	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if(rule_group == (*iter).name)
		{
			for(item_iter = (*iter).rule_items.begin();item_iter != (*iter).rule_items.end();++item_iter)
			{
				if(rule_name.compare((*item_iter).name,Qt::CaseInsensitive) == 0)
				{
					item = (*item_iter);
					
					if(valid)
					{
						*valid = true;
					}

					break;
				}
			}
			break;
		}
	}
	
	d_last_error = QStringLiteral("Rule not found");

	return item;
}

std::vector<TRules::TRuleItem> TRules::Rules(
	const QString						&rule_group) const
{
	std::vector<TRules::TRuleItem>		rules;
	std::vector<TRules::TRule>::const_iterator iter;

	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if(rule_group == (*iter).name)
		{
			rules = (*iter).rule_items;
			break;
		}
	}
	
	return rules;
}


void TRules::Set_Rule_Path(
	const QString						&path)
{
	d_rule_path = path;
}

bool TRules::Load_Rules(void)
{
	TXmlFile							xmlfile;
	int									version;
	
	try
	{
		xmlfile.SetFileName(QString("%1/rules.xml").arg(d_rule_path));
		xmlfile.Open("Rules",&version);
		
		switch(version)
		{
			case 1:
				this->Load_Version1(&xmlfile);
				break;
		}
		
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = xml_exception.ErrorText();
		return false;
	}
	
	this->Sort_Rules();
	
	d_data_changed = false;
	
	return true;
}

bool TRules::Create_Group(
	const QString						&group_name)
{
	std::vector<TRules::TRule>::iterator iter;
	TRules::TRule						rule;

	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if(group_name.compare((*iter).name,Qt::CaseInsensitive) == 0)
		{
			d_last_error = QStringLiteral("Group name already exists");
			return false;
		}
	}
	
	rule.name = group_name;
	d_rules.push_back(rule);
	
	d_data_changed = true;

	return true;
}

void TRules::Remove_Group(
	const QString						&group_name)
{
	std::vector<TRules::TRule>::iterator iter;
	
	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if(group_name.compare((*iter).name,Qt::CaseInsensitive) == 0)
		{
			d_rules.erase(iter);
			d_data_changed = true;
			
			return;
		}
	}
}

bool TRules::Add_Rule(
	const QString						&rule_group,
	const TRules::TRuleItem				&rule_item)
{
	std::vector<TRules::TRule>::iterator iter;
	std::vector<TRules::TRuleItem>::iterator item_iter;
	
	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if((*iter).name == rule_group)
		{
			for(item_iter = (*iter).rule_items.begin();item_iter != (*iter).rule_items.end();++item_iter)
			{
				if((*item_iter).name.compare(rule_item.name,Qt::CaseInsensitive) == 0)
				{
					d_last_error = QStringLiteral("Rule item already exists");
					return false;
				}
			}
			
			(*iter).rule_items.push_back(rule_item);
			d_data_changed = true;
			
			return true;
		}
	}
	
	d_last_error = QStringLiteral("Source rule group not found");
	return false;
}

void TRules::Delete_Rule(
	const QString						&rule_group,
	const QString						&rule_name)
{
	std::vector<TRules::TRule>::iterator iter;
	std::vector<TRules::TRuleItem>::iterator item_iter;

	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if((*iter).name == rule_group)
		{
			for(item_iter = (*iter).rule_items.begin();item_iter != (*iter).rule_items.end();++item_iter)
			{
				if((*item_iter).name.compare(rule_name,Qt::CaseInsensitive) == 0)
				{
					(*iter).rule_items.erase(item_iter);
					d_data_changed = true;

					return;
				}
			}
			
			break;
		}
	}
}

bool TRules::Copy_Rule(
	const QString						&source_rule_group,
	const QString						&destination_rule_group,
	const QString						&rule_name,
	const bool							overwrite_existing)
{
	std::vector<TRules::TRule>::iterator iter;
	std::vector<TRules::TRuleItem>::iterator item_iter;
	TRules::TRuleItem					item;
	
	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		if(source_rule_group.compare((*iter).name,Qt::CaseInsensitive) == 0)
		{
			for(item_iter = (*iter).rule_items.begin();item_iter != (*iter).rule_items.end();++item_iter)
			{
				if(rule_name.compare((*item_iter).name,Qt::CaseInsensitive) == 0)
				{
					item = (*item_iter);
					
					if(overwrite_existing)
					{
						this->Delete_Rule(destination_rule_group,rule_name);
					}
					
					return this->Add_Rule(destination_rule_group,item);
				}
			}
			
			break;
		}
	}
	
	d_last_error = QStringLiteral("Rule group not found");
	return false;
}

bool TRules::Write(void) const
{
	TXmlFile							xmlfile;
	QDomElement							en1;
	QDomElement							en2;
	QDomElement							en3;
	std::vector<TRule>::const_iterator iter;
	std::vector<TRuleItem>::const_iterator item_iter;
	QString								text;
	
	try
	{
		xmlfile.SetFileName(QString("%1/rules.xml").arg(d_rule_path));
		xmlfile.Reset("Rules",1);
		
		for(iter = d_rules.begin();iter != d_rules.end();++iter)
		{
			en1 = xmlfile.CreateNode(xmlfile.RootNode(),QStringLiteral("Rule_Group"));
			
			en2 = xmlfile.CreateNode(&en1,QStringLiteral("Name"));
			xmlfile.WriteTextNode(&en2,(*iter).name);
						
			for(item_iter = (*iter).rule_items.begin();item_iter != (*iter).rule_items.end();++item_iter)
			{
				en2 = xmlfile.CreateNode(&en1,QStringLiteral("Rule_Item"));
				
				en3 = xmlfile.CreateNode(&en2,QStringLiteral("Name"));
				xmlfile.WriteTextNode(&en3,(*item_iter).name);
				
				text = (*item_iter).left_expression.simplified();
				
				if(text.length())
				{
					en3 = xmlfile.CreateNode(&en2,QStringLiteral("Left_Expression"));
					xmlfile.WriteTextNode(&en3,text);
				}
				
				text = (*item_iter).right_expression.simplified();
				
				if(text.length())
				{
					en3 = xmlfile.CreateNode(&en2,QStringLiteral("Right_Expression"));
					xmlfile.WriteTextNode(&en3,text);
				}
				
				en3 = xmlfile.CreateNode(&en2,QStringLiteral("Operator"));

				switch((*item_iter).rule_operator)
				{
					case TRules::OPERATOR_LESS_THAN:
						xmlfile.WriteTextNode(&en3,QStringLiteral("OPERATOR_LESS_THAN"));
						break;
					case TRules::OPEATOR_EQUAL:
						xmlfile.WriteTextNode(&en3,QStringLiteral("OPEATOR_EQUAL"));
						break;
					case TRules::OPERATOR_GREATER_THAN:
						xmlfile.WriteTextNode(&en3,QStringLiteral("OPERATOR_GREATER_THAN"));
						break;
					case TRules::OPERATOR_NOT_LESS_THAN:
						xmlfile.WriteTextNode(&en3,QStringLiteral("OPERATOR_NOT_LESS_THAN"));
						break;
					case TRules::OPERATOR_NOT_EQUAL:
						xmlfile.WriteTextNode(&en3,QStringLiteral("OPERATOR_NOT_EQUAL"));
						break;
					case TRules::OPERATOR_NOT_GREATER_THAN:
						xmlfile.WriteTextNode(&en3,QStringLiteral("OPERATOR_NOT_GREATER_THAN"));
						break;
				}
			}
		}
		xmlfile.WriteXmlFile();
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = xml_exception.ErrorText();
		return false;
	}
	
	d_data_changed = false;
	
	return true;
	
}

void TRules::Load_Version1(
	TXmlFile							* const xml_file)
{
	QDomElement							en1;
	QDomElement							en2;
	QDomElement							en3;
	QDomElement							en4;
	QString								text;
	TRules::TRule						rule;
	TRules::TRuleItem					rule_item;
	
	assert(xml_file);
	
	d_rules.clear();
	
	en1 = xml_file->GetNode(xml_file->RootNode(),QStringLiteral("Rule_Group"),true);
	while(!en1.isNull())
	{
		en2 = xml_file->GetNode(&en1,QStringLiteral("Name"));
		rule.name = xml_file->ReadTextNode(&en2);
		
		rule.rule_items.clear();
		
		en2 = xml_file->GetNode(&en1,QStringLiteral("Rule_Item"),true);
		while(!en2.isNull())
		{
			en3 = xml_file->GetNode(&en2,QStringLiteral("Name"));
			rule_item.name = xml_file->ReadTextNode(&en3);
			
			en3 = xml_file->GetNode(&en2,QStringLiteral("Left_Expression"),true);
			
			if(en3.isNull())
			{
				rule_item.left_expression = QString();
			}
			else
			{
				rule_item.left_expression = xml_file->ReadTextNode(&en3);
			}
			
			en3 = xml_file->GetNode(&en2,QStringLiteral("Right_Expression"),true);
			
			if(en3.isNull())
			{
				rule_item.right_expression = QString();
			}
			else
			{
				rule_item.right_expression = xml_file->ReadTextNode(&en3);
			}
			
			en3 = xml_file->GetNode(&en2,QStringLiteral("Operator"));
			text = xml_file->ReadTextNode(&en3);

			if(0 == text.compare(QStringLiteral("OPERATOR_LESS_THAN"),Qt::CaseInsensitive))
			{
				rule_item.rule_operator = TRules::OPERATOR_LESS_THAN;
			}
			else if(0 == text.compare(QStringLiteral("OPEATOR_EQUAL"),Qt::CaseInsensitive))
			{
				rule_item.rule_operator = TRules::OPEATOR_EQUAL;
			}
			else if(0 == text.compare(QStringLiteral("OPERATOR_GREATER_THAN"),Qt::CaseInsensitive))
			{
				rule_item.rule_operator = TRules::OPERATOR_GREATER_THAN;
			}
			else if(0 == text.compare(QStringLiteral("OPERATOR_NOT_LESS_THAN"),Qt::CaseInsensitive))
			{
				rule_item.rule_operator = TRules::OPERATOR_NOT_LESS_THAN;
			}
			else if(0 == text.compare(QStringLiteral("OPERATOR_NOT_EQUAL"),Qt::CaseInsensitive))
			{
				rule_item.rule_operator = TRules::OPERATOR_NOT_EQUAL;
			}
			else if(0 == text.compare(QStringLiteral("OPERATOR_NOT_GREATER_THAN"),Qt::CaseInsensitive))
			{
				rule_item.rule_operator = TRules::OPERATOR_NOT_GREATER_THAN;
			}
			else
			{
				throw TXmlFileException(QStringLiteral("Operator type not recognized"));
			}
			
			rule.rule_items.push_back(rule_item);
			
			en2 = xml_file->GetSiblingNode(&en2,QStringLiteral("Rule_Item"),true);
		}
		
		d_rules.push_back(rule);
		
		en1 = xml_file->GetSiblingNode(&en1,QStringLiteral("Rule_Group"),true);
	}
}

void TRules::Sort_Rules(void)
{
	std::vector<TRules::TRule>::iterator iter;
	TRuleItemSort						rule_item_sort;
	
	for(iter = d_rules.begin();iter != d_rules.end();++iter)
	{
		std::sort((*iter).rule_items.begin(),(*iter).rule_items.end(),rule_item_sort);
	}
}
