/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QGridLayout>
#include <QLabel>
#include <QLineEdit>
#include <QSpacerItem>
#include <QSize>
#include <cmath>
#include <assert.h>

#include "utilities.h"

#include "valueswidget.h"


TValuesWidget::TValuesWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QFrame(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QSpacerItem							*widget_hspacer;
	QSpacerItem							*widget_vspacer;
	TValueWidget						value_widget;
	
	widget_layout = new QGridLayout(this);
	d_values_layout = new QGridLayout();
	d_values_layout->setVerticalSpacing(1);
	
	value_widget.index_label = new QLabel(this);
	d_values_layout->addWidget(value_widget.index_label,0,0,1,1);
	
	value_widget.value_edit = new QLineEdit(this);
	value_widget.value_edit->setAlignment(Qt::AlignCenter);
	value_widget.value_edit->setMinimumSize(QSize(200,0));
	value_widget.index = 0;
	d_values_layout->addWidget(value_widget.value_edit,0,1,1,1);

	widget_layout->addLayout(d_values_layout,0,0,1,1);

	widget_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	widget_layout->addItem(widget_hspacer,0,1,2,1);

	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,1,0,1,1);
	
	
	this->setWindowTitle(QStringLiteral("Values Widget"));
	
	value_widget.index_label->setText(QStringLiteral("Index:"));

	
	this->setFrameShape(QFrame::StyledPanel);

	// defaults
	d_value_widgets.push_back(value_widget);
	
	connect(value_widget.value_edit,SIGNAL(editingFinished(void)),this,SLOT(Value_Changed(void)));
}

TValuesWidget::~TValuesWidget(void)
{
}

int TValuesWidget::Integer(
	const int							index) const
{
	return static_cast<int>(this->Float(index));
}

double TValuesWidget::Float(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_value_widgets.size());
	
	return d_value_widgets[index].value_edit->text().toDouble();
}

QString TValuesWidget::Text(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_value_widgets.size());
	
	return d_value_widgets[index].value_edit->text();
}

void TValuesWidget::Reset(
	const int							values_count)
{
	int									cntr;
	TValueWidget						value_widget;
	bool								prev_state;
	std::vector<TValuesWidget::TValueWidget>::iterator iter;
	
	assert(!(values_count < 0));
	
	prev_state = d_values_layout->blockSignals(true);
	
	if(static_cast<unsigned int>(values_count) > d_value_widgets.size())
	{
		for(cntr = static_cast<int>(d_value_widgets.size());cntr < values_count;++cntr)
		{
			value_widget.index_label = new QLabel(this);
			d_values_layout->addWidget(value_widget.index_label,cntr,0,1,1);

			value_widget.value_edit = new QLineEdit(this);
			value_widget.value_edit->setAlignment(Qt::AlignCenter);
			value_widget.index = cntr;

			d_values_layout->addWidget(value_widget.value_edit,cntr,1,1,1);
			
			d_value_widgets.push_back(value_widget);
			
			connect(value_widget.value_edit,SIGNAL(editingFinished(void)),this,SLOT(Value_Changed(void)));
		}
	}
	
	for(iter = d_value_widgets.begin(),cntr = 0;iter != d_value_widgets.end();++iter,++cntr)
	{
		if(cntr < values_count)
		{
			(*iter).index_label->setVisible(true);
			(*iter).value_edit->setVisible(true);
		}
		else
		{
			(*iter).index_label->setVisible(false);
			(*iter).value_edit->setVisible(false);
		}
	}
	
	d_values_layout->blockSignals(prev_state);

}

void TValuesWidget::Set_Integer(
	const int							index,
	const int							value)
{
	assert(static_cast<unsigned int>(index) < d_value_widgets.size());
	
	d_value_widgets[index].index_label->setText(QString("Index: %1").arg(index + 1));
	d_value_widgets[index].value_edit->setText(QString("%1").arg(value));
}

void TValuesWidget::Set_Float(
	const int							index,
	const double						&value)
{
	bool								high_precision;
	
	assert(static_cast<unsigned int>(index) < d_value_widgets.size());
	
	d_value_widgets[index].index_label->setText(QString("Index: %1").arg(index + 1));
	
	high_precision = Utilities::Use_High_Precision_Display(value);
	
	if(high_precision)
	{
		d_value_widgets[index].value_edit->setText(QString("%1").arg(value,0,'f',Utilities::DISPLAY_PRECISION_HIGH));
	}
	else
	{
		d_value_widgets[index].value_edit->setText(QString("%1").arg(value,0,'f',Utilities::DISPLAY_PRECISION_NORMAL));
	}
}

void TValuesWidget::Set_Text(
	const int							index,
	const QString						&text)
{
	assert(static_cast<unsigned int>(index) < d_value_widgets.size());
	
	d_value_widgets[index].index_label->setText(QString("Index: %1").arg(index + 1));
	d_value_widgets[index].value_edit->setText(text);
}

void TValuesWidget::Value_Changed(void)
{
	std::vector<TValueWidget>::const_iterator iter;
	QObject								*source;
	
	source = this->sender();
	
	if(source)
	{
		for(iter = d_value_widgets.begin();iter != d_value_widgets.end();++iter)
		{
			if(source == (*iter).value_edit)
			{
				emit Value_Changed((*iter).index);
				return;
			}
		}
	}
}


