/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QLineEdit>
#include <QEvent>
#include <QDropEvent>
#include <QMimeData>
#include <QDir>
#include <QString>
#include <QTimer>

#include "eventfilters.h"

static bool Drop_Target_Path(const QList<QUrl> &items,QObject *target);

TDragDropFileEventFilter::TDragDropFileEventFilter(
	const QObject						*parent)
:QObject(const_cast<QObject*>(parent))
{
	d_signal_timer = new QTimer;
	
	d_signal_timer->setSingleShot(true);
	connect(d_signal_timer,SIGNAL(timeout(void)),this,SIGNAL(Drop_Complete(void)));
}

TDragDropFileEventFilter::~TDragDropFileEventFilter(void)
{
	delete d_signal_timer;
}

bool TDragDropFileEventFilter::eventFilter(
	QObject								*object,
	QEvent								*event)
{
	QDropEvent							*drop_event;
	
	switch(event->type())
	{
		case QEvent::DragEnter:
			event->accept();
			return true;
			
		case QEvent::Drop:
			drop_event = reinterpret_cast<QDropEvent*>(event);
			
			if(drop_event)
			{
				const QMimeData *mime_data = drop_event->mimeData();
				
				if(mime_data->hasUrls())
				{
					if(Drop_Target_Path(mime_data->urls(),object))
					{
						event->accept();
						d_signal_timer->start(10);	// don't want to block the event loop needlessly
					}
					else
					{
						event->ignore();
					}

					return true;
				}
			}
			break;
			
		default:
			break;
	}
	
	return QObject::eventFilter(object,event);
}

TDragDropDirEventFilter::TDragDropDirEventFilter(
	const QObject						*parent)
:QObject(const_cast<QObject*>(parent))
{
	d_signal_timer = new QTimer;
	
	d_signal_timer->setSingleShot(true);
	connect(d_signal_timer,SIGNAL(timeout(void)),this,SIGNAL(Drop_Complete(void)));
}

TDragDropDirEventFilter::~TDragDropDirEventFilter(void)
{
	delete d_signal_timer;
}

bool TDragDropDirEventFilter::eventFilter(
   QObject								*object,
   QEvent								*event)
{
	QDropEvent							*drop_event;
	QList<QUrl>							urls;
	
	switch(event->type())
	{
		case QEvent::DragEnter:
			event->accept();
			return true;
			
		case QEvent::Drop:
			drop_event = reinterpret_cast<QDropEvent*>(event);
			
			if(drop_event)
			{
				const QMimeData *mime_data = drop_event->mimeData();
				
				if(mime_data->hasUrls())
				{
					urls = this->Convert_To_Directories(mime_data->urls());
					
					if(Drop_Target_Path(urls,object))
					{
						event->accept();
						d_signal_timer->start(10);	// don't want to block the event loop needlessly
					}
					else
					{
						event->ignore();
					}
					
					return true;
				}
			}
			break;
			
		default:
			break;
	}
	
	return QObject::eventFilter(object,event);
}

QList<QUrl> TDragDropDirEventFilter::Convert_To_Directories(
	const QList<QUrl>					&urls) const
{
	QList<QUrl>							dir_urls;
	QList<QUrl>::const_iterator			iter;
	QString								text;
	QDir								dir;
	
	for(iter = urls.begin();iter != urls.end();++iter)
	{
		text = (*iter).toLocalFile();
		dir.setPath(text);
		
		if(dir.exists())
		{
			dir_urls.push_back(*iter);
		}
	}
	
	return dir_urls;
}


bool Drop_Target_Path(
	const QList<QUrl>					&items,
	QObject								*target)
{
	QLineEdit							*line_edit;
	
	if(items.size())
	{
		line_edit = qobject_cast<QLineEdit*>(target);
		
		if(line_edit)
		{			
			line_edit->setText(QDir::toNativeSeparators(items[0].toLocalFile()));	// only use first URL
			return true;
		}
		
		// other widget types will be added as needed
	}

	return false;
}

