/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QtXml/QDomNodeList>
#include <QtXml/QDomText>
#include <QtXml/QDomNamedNodeMap>
#include <QFile>
#include <QTextStream>
#include <QFileInfo>
#include <assert.h>

#include "xmlfile.h"

TXmlFileException::TXmlFileException(
	const QString 						&Error,
	const TXmlFileException::TExceptionType	 Type)
:d_ErrorText(Error),
 d_ErrorType(Type)
{

}

TXmlFileException::~TXmlFileException(void)
{

}

TXmlFile::TXmlFile(void)
{
	QDomImplementation::setInvalidDataPolicy(QDomImplementation::ReturnNullNode);
}

TXmlFile::~TXmlFile(void)
{

}

void TXmlFile::Open(
	const QString						&FileIdentification,
	int 								*VersionNumber,
	const bool							process_namespace)
{
	QFile								File;
	QString								Error;
	int									ErrorLine;
	int									ErrorColumn;
	QString								Text;
	QDomDocumentType					XmlDataType;
	QFileInfo							FileInfo(d_FileName);

	if(!FileInfo.isReadable())
	{
		Text = QString("ERR:  Cannot open file '%1' for reading.").arg(d_FileName);
		throw TXmlFileException(Text,TXmlFileException::FILE_OPEN_ERROR);
	}

	File.setFileName(d_FileName);

	if(!File.open(QFile::ReadOnly | QFile::Text))
	{
		Text = QString("ERR:  Cannot open file: '%1'.  Reason: %2").arg(d_FileName).arg(File.errorString());
		throw TXmlFileException(Text,TXmlFileException::FILE_OPEN_ERROR);
	}

	if(!d_XmlData.setContent(
					&File,
					process_namespace,
					&Error,
					&ErrorLine,
					&ErrorColumn))
	{
		File.close();

		Text = QString("ERR: Reference file XML parse error at line %1 column %2: %3")
					.arg(Error)
					.arg(ErrorLine)
					.arg(ErrorColumn);
		throw TXmlFileException(Text);
	}

	File.close();

	d_RootNode = d_XmlData.documentElement();

	if(d_RootNode.tagName() != FileIdentification)
	{
		Text = QString("ERR:  Identification '%1' does not match '%2'")
									.arg(FileIdentification)
									.arg(d_RootNode.tagName());

		throw TXmlFileException(Text);
	}

	XmlDataType = d_XmlData.doctype();
	d_DocumentType = XmlDataType.name();

	d_FileIdentification = FileIdentification;

	if(VersionNumber)
	{
		if(d_RootNode.hasAttribute("File_Version"))
		{
			*VersionNumber = d_RootNode.attribute("File_Version").toInt();
		}
		else
		{
			Text = QString("ERR: The 'File_Version' attribute is missing.");
			throw TXmlFileException(Text);
		}
	}
}

QString TXmlFile::ReadTextNode(
	const QDomNode 						*Node,
	const bool 							Optional) const
{
	QString								Text;
	QDomNode							TextNode;

	assert(Node);

	if(Node->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot read data from a NULL node"));
	}

	TextNode = Node->firstChild();

	if(TextNode.isNull() == false && TextNode.isText() == true)
	{
		Text = TextNode.nodeValue();
	}
	else
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node Text for '%1' is not optional.").arg(Node->localName()));
		}

		Text = QString();
	}

	return Text;
}

QString TXmlFile::ReadTextNode(
	const QDomElement 					*Node,
	const bool 							Optional) const
{
	QString								Text;
	QDomNode							TextNode;

	assert(Node);

	if(Node->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot read data from a NULL node"));
	}

	TextNode = Node->firstChild();

	if(TextNode.isNull() == false && TextNode.isText() == true)
	{
		Text = TextNode.nodeValue();
	}
	else
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node Text for '%1' is not optional.").arg(Node->tagName()));
		}

		Text = QString();
	}

	return Text;
}

QString TXmlFile::ReadAttribute(
	const QDomElement 					*Node,
	const QString 						&Name,
	const bool 							Optional) const
{
	QString								Text;

	assert(Node);

	if(Node->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot read attribute from a NULL node"));
	}

	if(Node->hasAttribute(Name))
	{
		Text = Node->attribute(Name);
	}
	else
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node attribute '%1' missing and not optional.").arg(Name));
		}

		Text = QString();
	}

	return Text;
}

QString TXmlFile::ReadAttribute(
	const QDomNode 						*Node,
	const QString 						&Name,
	const bool 							Optional) const
{
	QDomNamedNodeMap					NamedNodeMap;
	QDomNode							AttributeNode;
	QString								Text;
	bool								FoundAttribute(false);

	assert(Node);

	if(Node->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot read attribute from a NULL node"));
	}

	if(Node->hasAttributes())
	{
		NamedNodeMap = Node->attributes();

		if(NamedNodeMap.contains(Name))
		{
			AttributeNode = NamedNodeMap.namedItem(Name);
			Text = AttributeNode.nodeValue();
			FoundAttribute = true;
		}
	}

	if(!FoundAttribute)
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node attribute '%1' missing and not optional.").arg(Name));
		}

		Text = QString();
	}

	return Text;
}

QDomElement TXmlFile::GetNode(
	const QDomElement 					*Node,
	const QString 						&Name,
	const bool 							Optional) const
{
	QDomElement							ChildNode;

	assert(Node);

	if(Node->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot get child from a NULL node"));
	}

	ChildNode = Node->firstChildElement(Name);

	if(ChildNode.isNull())
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node '%1' is missing from node '%2' and not optional.")
												.arg(Name)
												.arg(Node->tagName()));
		}
	}

	return ChildNode;
}

QDomElement TXmlFile::GetNode(
	const QDomNode 						*Node,
	const QString 						&Name,
	const bool 							Optional) const
{
	QDomElement							ChildNode;

	assert(Node);

	if(Node->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot get child from a NULL node"));
	}

	ChildNode = Node->firstChildElement(Name);

	if(ChildNode.isNull())
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node '%1' is missing from node '%2' and not optional.")
												.arg(Name)
												.arg(Node->nodeName()));
		}
	}

	return ChildNode;
}

QDomElement TXmlFile::GetSiblingNode(
	const QDomElement 					*CurrentNode,
	const QString 						&Name,
	const bool 							Optional) const
{
	QDomElement							NextNode;

	assert(CurrentNode);

	if(CurrentNode->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot get sibling from a NULL node"));
	}

	NextNode = CurrentNode->nextSiblingElement(Name);

	if(NextNode.isNull())
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node '%1' is missing from node '%2' and not optional.")
												.arg(Name)
												.arg(CurrentNode->tagName()));
		}
	}

	return NextNode;
}

QDomNode TXmlFile::GetFirstNode(
	const QDomElement 					*Parent,
	const bool 							Optional) const
{
	QDomNode							Node;

	assert(Parent);

	Node = Parent->firstChild();

	if(Node.isNull())
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Parent '%1' does not have nodes and they are not optional.")
												.arg(Parent->tagName()));
		}
	}

	return Node;
}

QDomNode TXmlFile::GetFirstNode(
	const QDomNode	 					*Parent,
	const bool 							Optional) const
{
	QDomNode							Node;

	assert(Parent);

	Node = Parent->firstChild();

	if(Node.isNull())
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Parent '%1' does not have nodes and they are not optional.")
												.arg(Parent->localName()));
		}
	}

	return Node;
}

QDomNode TXmlFile::GetSiblingNode(
	const QDomNode	 					*CurrentNode,
	const bool 							Optional) const
{
	QDomNode							Node;

	assert(CurrentNode);

	Node = CurrentNode->nextSibling();

	if(Node.isNull())
	{
		if(!Optional)
		{
			throw TXmlFileException(QString("ERR: Node '%1' does not have siblings and they are not optional.")
												.arg(CurrentNode->localName()));
		}
	}

	return Node;
}

QDomElement TXmlFile::CreateNode(
	QDomElement 						*Parent,
	const QString 						&Name)
{
	QDomElement							Node;

	assert(Parent);

	if(Parent->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot create node from a NULL node"));
	}

	if(Name.length() == 0)
	{
		throw TXmlFileException(QString("ERR:  Cannot create an unnamed node"));
	}

	Node = d_XmlData.createElement(Name);
	Parent->appendChild(Node);

	return Node;
}

void TXmlFile::Clear(void)
{
	d_XmlData.clear();
}

void TXmlFile::Reset(
	const QString 						&FileIdentification,
	const int 							VersionNumber)
{
	QDomProcessingInstruction 			ProcessInstruction;

	d_XmlData.clear();

    ProcessInstruction = d_XmlData.createProcessingInstruction("xml", "version=\"1.0\" encoding=\"utf-8\"");
    d_XmlData.appendChild(ProcessInstruction);

    d_RootNode = d_XmlData.createElement(FileIdentification);
    d_RootNode.setAttribute("File_Version",QString("%1").arg(VersionNumber));
    d_XmlData.appendChild(d_RootNode);
}

void TXmlFile::Reset(
	const QString 						&DocType,
	const QString 						&FileIdentification,
	const int 							VersionNumber)
{

	d_XmlData = QDomDocument(DocType);

    d_RootNode = d_XmlData.createElement(FileIdentification);
    d_RootNode.setAttribute("File_Version",QString("%1").arg(VersionNumber));
    d_XmlData.appendChild(d_RootNode);
}

void TXmlFile::WriteTextNode(
	QDomElement 						*Node,
	const QString 						&Data)
{
	QDomText							TextNode;

	assert(Node);

	if(Node->isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot append data to a NULL node"));
	}

	TextNode = d_XmlData.createTextNode(Data);

	if(TextNode.isNull())
	{
		throw TXmlFileException(QString("ERR:  Cannot write node '%1' with data '%2'")
														.arg(Node->tagName())
														.arg(Data));
	}
	Node->appendChild(TextNode);
}

void TXmlFile::WriteTextNode(
	QDomElement 						*Node,
	const double 						&Value,
	const int 							Precision)
{
	this->WriteTextNode(Node,QString("%1").arg(Value,0,'f',Precision));
}

void TXmlFile::WriteTextNode(
	QDomElement 						*Node,
	const int 							Value)
{
	this->WriteTextNode(Node,QString("%1").arg(Value));
}

void TXmlFile::WriteAttribute(
	QDomElement 						*Node,
	const QString						&Name,
	const QString 						&Attribute)
{
	assert(Node);

	if(Name.length() == 0)
	{
		throw TXmlFileException(QString("ERR:  Cannot create an unnamed attribute"));
	}

	Node->setAttribute(Name,Attribute);
}

void TXmlFile::WriteAttribute(
	QDomElement 						*Node,
	const QString 						&Name,
	const double 						&Attribute,
	const int							Precision)
{
	this->WriteAttribute(Node,Name,QString("%1").arg(Attribute,0,'f',Precision));
}

void TXmlFile::WriteAttribute(
	QDomElement 						*Node,
	const QString 						&Name,
	const int 							Attribute)
{
	this->WriteAttribute(Node,Name,QString("%1").arg(Attribute));
}


void TXmlFile::WriteXmlFile(
	const bool							inline_xml)
{
	QFile								File;
	const int							XmlIndentSize = 4;
	QString 							Text;
	QFileInfo							FileInfo(d_FileName);

	if(FileInfo.exists() == true && FileInfo.isWritable() == false)
	{
		Text = QString("ERR:  Cannot open file '%1' for writing").arg(d_FileName);
		throw TXmlFileException(Text,TXmlFileException::FILE_OPEN_ERROR);
	}

	File.setFileName(d_FileName);

    if (!File.open(QIODevice::WriteOnly | QIODevice::Text))
    {
		Text = QString("ERR:  Cannot write file '%1'.  Reason: %2").arg(d_FileName).arg(File.errorString());
		throw TXmlFileException(Text);
    }

	if(inline_xml)
	{
		QByteArray						byte_array;
		QByteArray						write_byte_array;
		int								cntr;
		QTextStream 					TextStream(&byte_array);
		
		d_XmlData.save(TextStream, 0);
		
		for(cntr = 0;cntr < byte_array.size();++cntr)
		{
			if(byte_array[cntr] != '\n' && byte_array[cntr] != '\r')
			{
				write_byte_array.push_back(byte_array[cntr]);
			}
		}
		
		File.write(write_byte_array);
	}
	else
	{
		QTextStream 					TextStream(&File);

		d_XmlData.save(TextStream, XmlIndentSize);
	}
	
	File.close();
}


